__all__ = ("hash_", "Caching", "cache", "CacheInfo", "CachedObject",)

import sys
import functools
from collections import OrderedDict
from inspect import signature
from _thread import RLock

import numpy as np


def _to_hashable(obj):
    """Make unhashable objects hashable in a consistent manner."""

    if isinstance(obj, (int, float, str)):
        # Strings and Numbers are hashed directly.
        return obj

    elif hasattr(obj, "__iter__"):
        # Encapsulate all the iterables to quickly discard as needed.

        if isinstance(obj, np.ndarray):
            # Numpy arrays: Convert the data buffer to a byte string.
            return obj.tobytes()

        elif isinstance(obj, dict):
            # Dictionaries: Build a tuple from key-value pairs,
            # where all values are converted to hashables.
            out = {key: _to_hashable(value) for key, value in obj.items()}
            # Sort unordered dictionaries for hash consistency.
            if isinstance(obj, OrderedDict):
                return tuple(out.items())
            return tuple(sorted(out.items()))

        else:
            # Iterables: Build a tuple from values converted to hashables.
            out = [_to_hashable(item) for item in obj]
            return tuple(out)

    elif hasattr(obj, "__hash__"):
        # Hashables: Just return the object.
        return obj

    # NotImplemented: Can't hash safely, so raise TypeError.
    # Note: This will never be triggered since `type` has a repr slot wrapper.
    raise TypeError(f"Hashing for {type(obj)} not implemented.")


def hash_(obj):
    """Generic hash method, which changes between processes."""
    digest = hash(repr(_to_hashable(obj))) + sys.maxsize + 1
    return digest


class _CachingMeta(type):
    """Implement ``property`` to a ``classmethod`` for ``Caching``."""
    # NOTE: Only in 3.8 < py < 3.11 can `classmethod` wrap `property`.
    # https://docs.python.org/3.11/library/functions.html#classmethod
    @property
    def maxsize(cls):
        return cls._maxsize

    @maxsize.setter
    def maxsize(cls, value):
        if value < 0:
            raise ValueError(
                "`maxsize` should be larger than zero. "
                "To disable caching, use `Caching.disable()`.")
        cls._maxsize = value
        for func in cls._cached_functions:
            func.cache_info.maxsize = value

    @property
    def policy(cls):
        return cls._policy

    @policy.setter
    def policy(cls, value):
        if value not in cls._policies:
            raise ValueError("Cache retention policy not recognized.")
        if value == "lfu" != cls._policy:
            # Reset counter if we change policy to lfu
            # otherwise new objects are prone to being discarded immediately.
            # Now, the counter is not just used for stats,
            # it is part of the retention policy.
            for func in cls._cached_functions:
                for item in func.cache_info._caches.values():
                    item.reset()
        cls._policy = value
        for func in cls._cached_functions:
            func.cache_info.policy = value


class Caching(metaclass=_CachingMeta):
    """Infrastructure to hold cached objects.

    Caching is used for pre-computed objects that are expensive to compute.

    Attributes:
        maxsize (``int``):
            Maximum number of caches to store. If the dictionary is full, new
            caches are assigned according to the set cache retention policy.
        policy (``'fifo'``, ``'lru'``, ``'lfu'``):
            Cache retention policy.
    """
    _enabled: bool = False
    _policies: list = ['fifo', 'lru', 'lfu']
    _default_maxsize: int = 128   # class default maxsize
    _default_policy: str = 'lru'  # class default policy
    _maxsize = _default_maxsize   # user-defined maxsize
    _policy = _default_policy     # user-defined policy
    _cached_functions: list = []

    @classmethod
    def _get_key(cls, func, *args, **kwargs):
        """Calculate the hex hash from arguments and keyword arguments."""
        # get a dictionary of default parameters
        params = func.cache_info._signature.parameters
        # get a dictionary of the passed parameters
        passed = {**dict(zip(params, args)), **kwargs}
        # discard the values equal to the default
        defaults = {param: value.default for param, value in params.items()}
        return hex(hash_({**defaults, **passed}))

    @classmethod
    def _get(cls, dic, key, policy):
        """Get the cached object container
        under the implemented caching policy.
        """
        obj = dic[key]
        if policy == "lru":
            dic.move_to_end(key)
        # update stats
        obj.increment()
        return obj

    @classmethod
    def _pop(cls, dic, policy):
        """Remove one cached item as per the implemented caching policy."""
        if policy == "lfu":
            keys = list(dic)
            idx = np.argmin([item.counter for item in dic.values()])
            dic.move_to_end(keys[idx], last=False)
        dic.popitem(last=False)

    @classmethod
    def _decorator(cls, func, maxsize, policy):
        # assign caching attributes to decorated function
        func.cache_info = CacheInfo(func, maxsize=maxsize, policy=policy)
        func.clear_cache = func.cache_info._clear_cache
        cls._cached_functions.append(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not cls._enabled:
                return func(*args, **kwargs)

            key = cls._get_key(func, *args, **kwargs)
            # shorthand access
            caches = func.cache_info._caches
            maxsize = func.cache_info.maxsize
            policy = func.cache_info.policy

            with RLock():
                if key in caches:
                    # output has been cached; update stats and return it
                    out = cls._get(caches, key, policy)
                    func.cache_info.hits += 1
                    return out.item

            with RLock():
                while len(caches) >= maxsize:
                    # output not cached and no space available, so remove
                    # items as per the caching policy until there is space
                    cls._pop(caches, policy)

            # cache new entry and update stats
            out = CachedObject(func(*args, **kwargs))
            caches[key] = out
            func.cache_info.misses += 1
            return out.item

        return wrapper

    @classmethod
    def cache(cls, func=None, *, maxsize=_maxsize, policy=_policy):
        """Cache the output of the decorated function, using the input
        arguments as a proxy to build a hash key.

        Arguments:
            func (``function``):
                Function to be decorated.
            maxsize (``int``):
                Maximum cache size for the decorated function.
            policy (``'fifo'``, ``'lru'``, ``'lfu'``):
                Cache retention policy. When the storage reaches maxsize
                decide which cached object will be deleted. Default is 'lru'.\n
                'fifo': first-in-first-out,\n
                'lru': least-recently-used,\n
                'lfu': least-frequently-used.
        """
        if maxsize < 0:
            raise ValueError(
                "`maxsize` should be larger than zero. "
                "To disable caching, use `Caching.disable()`.")
        if policy not in cls._policies:
            raise ValueError("Cache retention policy not recognized.")

        if func is None:
            # `@cache` with parentheses
            return functools.partial(
                cls._decorator, maxsize=maxsize, policy=policy)
        # `@cache()` without parentheses
        return cls._decorator(func, maxsize=maxsize, policy=policy)

    @classmethod
    def enable(cls):
        cls._enabled = True

    @classmethod
    def disable(cls):
        cls._enabled = False

    @classmethod
    def reset(cls):
        cls.maxsize = cls._default_maxsize
        cls.policy = cls._default_policy

    @classmethod
    def clear_cache(cls):
        [func.clear_cache() for func in cls._cached_functions]


cache = Caching.cache


class CacheInfo:
    """Cache info container.
    Assigned to cached function as ``function.cache_info``.

    Parameters:
        func (``function``):
            Function in which an instance of this class will be assigned.
        maxsize (``Caching.maxsize``):
            Maximum number of caches to store.
        policy (``Caching.policy``):
            Cache retention policy.

    .. note:: To assist in deciding an optimal ``maxsize`` and ``policy``,
              instances of this class contain the following attributes:
              - ``hits``: number of times the function has been bypassed.
              - ``misses``: number of times the function has computed
              something.
              - ``current_size``: current size of the cache dictionary.
    """

    def __init__(self, func, maxsize=Caching.maxsize, policy=Caching.policy):
        # we store the signature of the function on import
        # as it is the most expensive operation (~30x slower)
        self._signature = signature(func)
        self._caches = OrderedDict()
        self.maxsize = maxsize
        self.policy = policy
        self.hits = self.misses = 0

    @property
    def current_size(self):
        return len(self._caches)

    def __repr__(self):
        s = f"<{self.__class__.__name__}>"
        for par, val in self.__dict__.items():
            if not par.startswith("_"):
                s += f"\n\t {par} = {val!r}"
        s += f"\n\t current_size = {self.current_size!r}"
        return s

    def _clear_cache(self):
        self._caches = OrderedDict()
        self.hits = self.misses = 0


class CachedObject:
    """A cached object container.

    Attributes:
        counter (``int``):
            Number of times the cached item has been retrieved.
    """
    counter: int = 0

    def __init__(self, obj):
        self.item = obj

    def __repr__(self):
        s = f"CachedObject(counter={self.counter})"
        return s

    def increment(self):
        self.counter += 1

    def reset(self):
        self.counter = 0



import numpy as np
import matplotlib.pyplot as plt
import py_cosmo_mad as csm

TCMB=2.725
PLOT_STUFF=0
WRITE_STUFF=1
FS=16
LKMAX=7

def do_all(z_arr,k_arr,cpar,prefix,wig):
    pcs=csm.PcsPar()
    pcs.background_set(cpar['om'],cpar['ol'],cpar['ob'],cpar['w0'],cpar['wa'],cpar['hh'],TCMB)
    if wig:
        typ = 'EH'
    else:
        typ = 'EH_smooth'
    pcs.set_linear_pk(typ,-3,LKMAX,0.01,cpar['ns'],cpar['s8'])

    gf0=pcs.growth_factor(1)
    a_arr=1./(z_arr+1)
    gf_arr=np.array([pcs.growth_factor(a) for a in a_arr])
    pk_arr=np.array([[pcs.Pk_linear_0(k)*(gf/gf0)**2 for k in k_arr] for gf in gf_arr])

    if PLOT_STUFF==1 :
        for i in np.arange(len(pk_arr)) :
            plt.plot(k_arr,pk_arr[i],label='$z=%.2lf$'%(z_arr[i]))
        plt.xlabel('$k\\,[h\\,{\\rm Mpc}^{-1}]$',fontsize=FS)
        plt.ylabel('$P(k)\\,[{\\rm Mpc}\\,h^{-1}]^3$',fontsize=FS)
        plt.gca().set_xscale('log');
        plt.gca().set_yscale('log');
        plt.legend(loc='lower left')
        plt.show()

    if WRITE_STUFF==1 :
        header_pk="[0] k (Mpc/h)^-1"
        for i in np.arange(len(z_arr)) :
            header_pk+=", [%d]"%(i+2)+" P(k,z=%.1lf) (Mpc/h)^3"%(z_arr[i])
        np.savetxt(prefix+"_pk_eh.txt",np.transpose(np.vstack((k_arr,pk_arr))),header=header_pk)

z_arr=np.array([0.,1.,2.,3.,4.,5.])
lk_arr=0.1*np.arange(41)-3.
k_arr=10**lk_arr

cpar_model1={'om': 0.3,'ol': 0.7,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -1.0, 'wa': 0.0}

do_all(z_arr,k_arr,cpar_model1,"model1",True)
do_all(z_arr,k_arr,cpar_model1,"model1_nowig",False)



__all__ = ("MassFuncTinker10",)

import numpy as np
from scipy.interpolate import interp1d

from . import MassFunc, get_delta_c


class MassFuncTinker10(MassFunc):
    """Implements the mass function of `Tinker et al. 2010
    <https://arxiv.org/abs/1001.3162>`_. This parametrization accepts S.O.
    masses with :math:`200 < \\Delta < 3200`, defined with respect to the
    matter density. This can be automatically translated to S.O. masses
    defined with respect to the critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
        norm_all_z (:obj:`bool`): if ``True``, the mass function will be
            normalised to yield the total matter density when integrated
            over mass at all redshifts (as opposed to :math:`z=0` only).
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict",
                                     "norm_all_z",)
    name = 'Tinker10'

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True,
                 norm_all_z=False):
        self.norm_all_z = norm_all_z
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta == "fof"

    def _setup(self):
        delta = np.array(
            [200., 300., 400., 600., 800., 1200., 1600., 2400., 3200.])
        alpha = np.array(
            [0.368, 0.363, 0.385, 0.389, 0.393, 0.365, 0.379, 0.355, 0.327])
        beta = np.array(
            [0.589, 0.585, 0.544, 0.543, 0.564, 0.623, 0.637, 0.673, 0.702])
        gamma = np.array(
            [0.864, 0.922, 0.987, 1.09, 1.20, 1.34, 1.50, 1.68, 1.81])
        phi = np.array(
            [-0.729, -0.789, -0.910, -1.05, -1.20, -1.26, -1.45, -1.50, -1.49])
        eta = np.array(
            [-0.243, -0.261, -0.261, -0.273,
             -0.278, -0.301, -0.301, -0.319, -0.336])

        ldelta = np.log10(delta)
        self.pA0 = interp1d(ldelta, alpha)
        self.pa0 = interp1d(ldelta, eta)
        self.pb0 = interp1d(ldelta, beta)
        self.pc0 = interp1d(ldelta, gamma)
        self.pd0 = interp1d(ldelta, phi)
        if self.norm_all_z:
            p = np.array(
                [-0.158, -0.195, -0.213, -0.254, -0.281,
                 -0.349, -0.367, -0.435, -0.504])
            q = np.array(
                [0.0128, 0.0128, 0.0143, 0.0154, 0.0172,
                 0.0174, 0.0199, 0.0203, 0.0205])
            self.pp0 = interp1d(ldelta, p)
            self.pq0 = interp1d(ldelta, q)

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        ld = np.log10(self.mass_def._get_Delta_m(cosmo, a))
        nu = get_delta_c(cosmo, a, 'EdS_approx') / sigM
        # redshift evolution only up to z=3
        a = np.clip(a, 0.25, 1)
        pa = self.pa0(ld) * a**(-0.27)
        pb = self.pb0(ld) * a**(-0.20)
        pc = self.pc0(ld) * a**0.01
        pd = self.pd0(ld) * a**0.08
        pA0 = self.pA0(ld)
        if self.norm_all_z:
            z = 1/a - 1
            pp = self.pp0(ld)
            pq = self.pq0(ld)
            pA0 *= np.exp(z * (pp + pq * z))
        return nu * pA0 * (1 + (pb * nu)**(-2 * pd)) * (
            nu**(2 * pa) * np.exp(-0.5 * pc * nu**2))



__all__ = ("HaloProfile", "HaloProfileMatter",
           "HaloProfilePressure", "HaloProfileCIB",)

import functools
from typing import Callable

import numpy as np

from ... import CCLAutoRepr, FFTLogParams, unlock_instance
from ... import physical_constants as const
from ...pyutils import resample_array, _fftlog_transform
from .. import MassDef


class HaloProfile(CCLAutoRepr):
    """ This class implements functionality associated to
    halo profiles. You should not use this class directly.
    Instead, use one of the subclasses implemented in CCL
    for specific halo profiles, or write your own subclass.
    ``HaloProfile`` classes contain methods to compute halo
    profiles in real (3D) and Fourier spaces, as well as other
    projected (2D) quantities.

    A minimal implementation of a ``HaloProfile`` subclass
    should contain a method ``_real`` that returns the
    real-space profile as a function of cosmology,
    comoving radius, mass and scale factor. The default
    functionality in the base ``HaloProfile`` class will then
    allow the automatic calculation of the Fourier-space
    and projected profiles, as well as the cumulative
    mass density, based on the real-space profile using
    FFTLog to carry out fast Hankel transforms. See the
    CCL note for details. Alternatively, you can implement
    a ``_fourier`` method for the Fourier-space profile, and
    all other quantities will be computed from it. It is
    also possible to implement specific versions of any
    of these quantities if one wants to avoid the FFTLog
    calculation.
    """

    def __init__(self, *, mass_def, concentration=None,
                 is_number_counts=False):
        # Verify that profile can be initialized.
        if not (hasattr(self, "_real") or hasattr(self, "_fourier")):
            name = type(self).__name__
            raise TypeError(f"Can't instantiate {name} with no "
                            "_real or _fourier implementation.")

        # Initialize FFTLog.
        self.precision_fftlog = FFTLogParams()

        self._is_number_counts = is_number_counts

        # Initialize mass_def and concentration.
        self.mass_def, *out = MassDef.from_specs(
            mass_def, concentration=concentration)
        if out:
            self.concentration = out[0]

    @property
    def is_number_counts(self):
        """If ``True``, this profile represents source-count overdensities,
        normalised by the mean number density within their survey window
        function. This must be propagated when estimated super-sample
        effects in the covariance matrix.
        """
        return self._is_number_counts

    @is_number_counts.setter
    @unlock_instance
    def is_number_counts(self, value):
        self._is_number_counts = value

    def get_normalization(self, cosmo=None, a=None, *, hmc=None):
        """Profiles may be normalized by an overall function of redshift
        (or scale factor). This function may be cosmology-dependent and
        often comes from integrating certain halo properties over mass.
        This method returns this normalizing factor. For example,
        to get the normalized profile in real space, one would call
        the :meth:`real` method, and then **divide** the result by the value
        returned by this method.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.
            hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo
                model calculator object.

        Returns:
            float: normalization factor of this profile.
        """
        return 1.0

    @unlock_instance(mutate=True)
    @functools.wraps(FFTLogParams.update_parameters)
    def update_precision_fftlog(self, **kwargs):
        self.precision_fftlog.update_parameters(**kwargs)

    def _get_plaw_fourier(self, cosmo, a):
        """ This controls the value of `plaw_fourier` to be used
        as a function of cosmology and scale factor.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.

        Returns:
            float: power law index to be used with FFTLog.
        """
        return self.precision_fftlog['plaw_fourier']

    def _get_plaw_projected(self, cosmo, a):
        """ This controls the value of `plaw_projected` to be
        used as a function of cosmology and scale factor.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.

        Returns:
            float: power law index to be used with FFTLog.
        """
        return self.precision_fftlog['plaw_projected']

    _real: Callable       # implementation of the real profile

    _fourier: Callable    # implementation of the Fourier profile

    _projected: Callable  # implementation of the projected profile

    _cumul2d: Callable    # implementation of the cumulative surface density

    def real(self, cosmo, r, M, a):
        """
        real(cosmo, r, M, a)
        Returns the 3D real-space value of the profile as a
        function of cosmology, radius, halo mass and scale factor.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            r (:obj:`float` or `array`): comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): halo profile. The shape of the
            output will be `(N_M, N_r)` where `N_r` and `N_m` are
            the sizes of `r` and `M` respectively. If `r` or `M`
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """
        if getattr(self, "_real", None):
            return self._real(cosmo, r, M, a)
        return self._fftlog_wrap(cosmo, r, M, a, fourier_out=False)

    def fourier(self, cosmo, k, M, a):
        """
        fourier(cosmo, k, M, a)
        Returns the Fourier-space value of the profile as a
        function of cosmology, wavenumber, halo mass and
        scale factor.

        .. math::
           \\rho(k)=\\frac{1}{2\\pi^2} \\int dr\\, r^2\\,
           \\rho(r)\\, j_0(k r)

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber (in :math:`Mpc^{-1}`).
            M (:obj:`float` or `array`): halo mass.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): Fourier-space profile. The shape of the
            output will be ``(N_M, N_k)`` where ``N_k`` and ``N_m`` are
            the sizes of ``k`` and ``M`` respectively. If ``k`` or ``M``
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """ # noqa
        if getattr(self, "_fourier", None):
            return self._fourier(cosmo, k, M, a)
        return self._fftlog_wrap(cosmo, k, M, a, fourier_out=True)

    def projected(self, cosmo, r_t, M, a):
        """
        projected(cosmo, r_t, M, a)
        Returns the 2D projected profile as a function of
        cosmology, radius, halo mass and scale factor.

        .. math::
           \\Sigma(R)= \\int dr_\\parallel\\,
           \\rho(\\sqrt{r_\\parallel^2 + R^2})

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            r_t (:obj:`float` or `array`): transverse comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): projected profile. The shape of the
            output will be `(N_M, N_r)` where `N_r` and `N_m` are
            the sizes of `r` and `M` respectively. If `r` or `M`
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """
        if getattr(self, "_projected", None):
            return self._projected(cosmo, r_t, M, a)
        return self._projected_fftlog_wrap(cosmo, r_t, M, a, is_cumul2d=False)

    def cumul2d(self, cosmo, r_t, M, a):
        """
        cumul2d(cosmo, r_t, M, a)
        Returns the 2D cumulative surface density as a
        function of cosmology, radius, halo mass and scale
        factor.

        .. math::
           \\Sigma(<R)= \\frac{2}{R^2} \\int dR'\\, R'\\,
           \\Sigma(R')

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            r_t (:obj:`float` or `array`): transverse comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a (:obj:`float`): scale factor.

        Returns:
            (:obj:`float` or `array`): cumulative surface density. The shape of the
            output will be ``(N_M, N_r)`` where ``N_r`` and ``N_m`` are
            the sizes of ``r`` and ``M`` respectively. If ``r`` or ``M``
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """ # noqa
        if getattr(self, "_cumul2d", None):
            return self._cumul2d(cosmo, r_t, M, a)
        return self._projected_fftlog_wrap(cosmo, r_t, M, a, is_cumul2d=True)

    def convergence(self, cosmo, r, M, *, a_lens, a_source):
        """
        convergence(cosmo, r, M, *, a_lens, a_source)
        Returns the convergence as a function of cosmology,
        radius, halo mass and the scale factors of the source
        and the lens.

        .. math::
           \\kappa(R) = \\frac{\\Sigma(R)}{\\Sigma_{\\rm crit}},

        where :math:`\\Sigma(R)` is the 2D projected surface mass density
        (see :meth:`projected`), and :math:`\\Sigma_{\\rm crit}` is
        the critical surface density (see
        :meth:`~pyccl.background.sigma_critical`).

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            r (:obj:`float` or `array`): comoving radius.
            M (:obj:`float` or `array`): halo mass.
            a_lens (:obj:`float`): scale factor of lens.
            a_source (:obj:`float` or `array`): scale factor of source.
                If an array, it must have the same shape as ``r``.

        Returns:
            (:obj:`float` or `array`): convergence :math:`\\kappa`. The
            shape of the output will be ``(N_M, N_r)`` where ``N_r`` and
            ``N_m`` are the sizes of ``r`` and ``M`` respectively. If
            ``r`` or ``M`` are scalars, the corresponding dimension will
            be squeezed out on output.
        """
        Sigma = self.projected(cosmo, r, M, a_lens)
        Sigma /= a_lens**2
        Sigma_crit = cosmo.sigma_critical(a_lens=a_lens, a_source=a_source)
        return Sigma / Sigma_crit

    def shear(self, cosmo, r, M, *, a_lens, a_source):
        """
        shear(cosmo, r, M, *, a_lens, a_source)
        Returns the shear (tangential) as a function of cosmology,
        radius, halo mass and the scale factors of the
        source and the lens.

        .. math::
           \\gamma(R) = \\frac{\\Delta\\Sigma(R)}{\\Sigma_{\\mathrm{crit}}} =
           \\frac{\\overline{\\Sigma}(< R) -
           \\Sigma(R)}{\\Sigma_{\\mathrm{crit}}},

        where :math:`\\overline{\\Sigma}(< R)` is cumulative surface density
        (see :meth:`cumul2d`), :math:`\\Sigma(R)` is the projected 2D density
        (see :meth:`projected`), and :math:`\\Sigma_{\\rm crit}` is the
        critical surface density (see
        :meth:`~pyccl.background.sigma_critical`).

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            r (:obj:`float` or `array`): comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a_lens (:obj:`float`): scale factor of lens.
            a_source (:obj:`float` or `array`): source's scale factor.
                If array, it must have the same shape as `r`.

        Returns:
            (:obj:`float` or `array`): shear :math:`\\gamma`. The shape of the
            output will be ``(N_M, N_r)`` where ``N_r`` and ``N_m`` are
            the sizes of ``r`` and ``M`` respectively. If ``r`` or ``M``
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """
        Sigma = self.projected(cosmo, r, M, a_lens)
        Sigma_bar = self.cumul2d(cosmo, r, M, a_lens)
        Sigma_crit = cosmo.sigma_critical(a_lens=a_lens, a_source=a_source)
        return (Sigma_bar - Sigma) / (Sigma_crit * a_lens**2)

    def reduced_shear(self, cosmo, r, M, *, a_lens, a_source):
        """
        reduced_shear(cosmo, r, M, *, a_lens, a_source)
        Returns the reduced shear as a function of cosmology,
        radius, halo mass and the scale factors of the
        source and the lens.

        .. math::
           g_t (R) = \\frac{\\gamma(R)}{(1 - \\kappa(R))},

        where :math:`\\gamma(R)` is the shear and :math:`\\kappa(R)` is the
        convergence.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            r (:obj:`float` or `array`): comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a_lens (:obj:`float`): scale factor of lens.
            a_source (:obj:`float` or `array`): source's scale factor.
                If array, it must have the same shape as `r`.

        Returns:
            (:obj:`float` or `array`): reduced shear :math:`g_t`. The shape
            of the output will be ``(N_M, N_r)`` where ``N_r`` and ``N_m``
            are the sizes of ``r`` and ``M`` respectively. If ``r`` or
            ``M`` are scalars, the corresponding dimension will be
            squeezed out on output.
        """
        convergence = self.convergence(cosmo, r, M, a_lens=a_lens,
                                       a_source=a_source)
        shear = self.shear(cosmo, r, M, a_lens=a_lens, a_source=a_source)
        return shear / (1.0 - convergence)

    def magnification(self, cosmo, r, M, *, a_lens, a_source):
        """
        magnification(cosmo, r, M, a_lens, a_source)
        Returns the magnification for input parameters.

        .. math::
           \\mu(R) = \\frac{1}{\\left[(1 - \\kappa(R))^2 -
           \\vert \\gamma(R) \\vert^2 \\right]]},

        where :math:`\\gamma(R)` is the shear and :math:`\\kappa(R)` is the
        convergence (see :meth:`shear` and :meth:`convergence`).

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
            r (:obj:`float` or `array`): comoving radius in Mpc.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a_lens (:obj:`float`): scale factor of lens.
            a_source (:obj:`float` or `array`): source's scale factor.
                If array, it must have the same shape as `r`.

        Returns:
            (:obj:`float` or `array`): magnification :math:`\\mu`. The shape
            of the output will be ``(N_M, N_r)`` where ``N_r`` and ``N_m``
            are the sizes of ``r`` and ``M`` respectively. If ``r`` or ``M``
            are scalars, the corresponding dimension will be
            squeezed out on output.
        """
        convergence = self.convergence(cosmo, r, M, a_lens=a_lens,
                                       a_source=a_source)
        shear = self.shear(cosmo, r, M, a_lens=a_lens, a_source=a_source)

        return 1.0 / ((1.0 - convergence)**2 - np.abs(shear)**2)

    def _fftlog_wrap(self, cosmo, k, M, a,
                     fourier_out=False,
                     large_padding=True, ell=0):
        # This computes the 3D Hankel transform
        #  \rho(k) = 4\pi \int dr r^2 \rho(r) j_ell(k r)
        # if fourier_out == True, and
        #  \rho(r) = \frac{1}{2\pi^2} \int dk k^2 \rho(k) j_ell(k r)
        # otherwise.

        # Select which profile should be the input
        if fourier_out:
            p_func = self._real
        else:
            p_func = self._fourier
        k_use = np.atleast_1d(k)
        M_use = np.atleast_1d(M)
        lk_use = np.log(k_use)
        nM = len(M_use)

        # k/r ranges to be used with FFTLog and its sampling.
        if large_padding:
            k_min = self.precision_fftlog['padding_lo_fftlog'] * np.amin(k_use)
            k_max = self.precision_fftlog['padding_hi_fftlog'] * np.amax(k_use)
        else:
            k_min = self.precision_fftlog['padding_lo_extra'] * np.amin(k_use)
            k_max = self.precision_fftlog['padding_hi_extra'] * np.amax(k_use)
        n_k = (int(np.log10(k_max / k_min)) *
               self.precision_fftlog['n_per_decade'])
        r_arr = np.geomspace(k_min, k_max, n_k)

        p_k_out = np.zeros([nM, k_use.size])
        # Compute real profile values
        p_real_M = p_func(cosmo, r_arr, M_use, a)
        # Power-law index to pass to FFTLog.
        plaw_index = self._get_plaw_fourier(cosmo, a)

        # Compute Fourier profile through fftlog
        k_arr, p_fourier_M = _fftlog_transform(r_arr, p_real_M,
                                               3, ell, plaw_index)
        lk_arr = np.log(k_arr)

        for im, p_k_arr in enumerate(p_fourier_M):
            # Resample into input k values
            p_fourier = resample_array(lk_arr, p_k_arr, lk_use,
                                       self.precision_fftlog['extrapol'],
                                       self.precision_fftlog['extrapol'],
                                       0, 0)
            p_k_out[im, :] = p_fourier
        if fourier_out:
            p_k_out *= (2 * np.pi)**3

        if np.ndim(k) == 0:
            p_k_out = np.squeeze(p_k_out, axis=-1)
        if np.ndim(M) == 0:
            p_k_out = np.squeeze(p_k_out, axis=0)
        return p_k_out

    def _projected_fftlog_wrap(self, cosmo, r_t, M, a, is_cumul2d=False):
        # This computes Sigma(R) from the Fourier-space profile as:
        # Sigma(R) = \frac{1}{2\pi} \int dk k J_0(k R) \rho(k)
        r_t_use = np.atleast_1d(r_t)
        M_use = np.atleast_1d(M)
        lr_t_use = np.log(r_t_use)
        nM = len(M_use)

        # k/r range to be used with FFTLog and its sampling.
        r_t_min = self.precision_fftlog['padding_lo_fftlog'] * np.amin(r_t_use)
        r_t_max = self.precision_fftlog['padding_hi_fftlog'] * np.amax(r_t_use)
        n_r_t = (int(np.log10(r_t_max / r_t_min)) *
                 self.precision_fftlog['n_per_decade'])
        k_arr = np.geomspace(r_t_min, r_t_max, n_r_t)

        sig_r_t_out = np.zeros([nM, r_t_use.size])
        # Compute Fourier-space profile
        if getattr(self, "_fourier", None):
            # Compute from `_fourier` if available.
            p_fourier = self._fourier(cosmo, k_arr, M_use, a)
        else:
            # Compute with FFTLog otherwise.
            lpad = self.precision_fftlog['large_padding_2D']
            p_fourier = self._fftlog_wrap(cosmo, k_arr, M_use, a,
                                          fourier_out=True, large_padding=lpad)
        if is_cumul2d:
            # The cumulative profile involves a factor 1/(k R) in
            # the integrand.
            p_fourier *= 2 / k_arr[None, :]

        # Power-law index to pass to FFTLog.
        if is_cumul2d:
            i_bessel = 1
            plaw_index = self._get_plaw_projected(cosmo, a) - 1
        else:
            i_bessel = 0
            plaw_index = self._get_plaw_projected(cosmo, a)

        # Compute projected profile through fftlog
        r_t_arr, sig_r_t_M = _fftlog_transform(k_arr, p_fourier,
                                               2, i_bessel,
                                               plaw_index)
        lr_t_arr = np.log(r_t_arr)

        if is_cumul2d:
            sig_r_t_M /= r_t_arr[None, :]
        for im, sig_r_t_arr in enumerate(sig_r_t_M):
            # Resample into input r_t values
            sig_r_t = resample_array(lr_t_arr, sig_r_t_arr,
                                     lr_t_use,
                                     self.precision_fftlog['extrapol'],
                                     self.precision_fftlog['extrapol'],
                                     0, 0)
            sig_r_t_out[im, :] = sig_r_t

        if np.ndim(r_t) == 0:
            sig_r_t_out = np.squeeze(sig_r_t_out, axis=-1)
        if np.ndim(M) == 0:
            sig_r_t_out = np.squeeze(sig_r_t_out, axis=0)
        return sig_r_t_out


class HaloProfileMatter(HaloProfile):
    """Base for matter halo profiles."""

    def get_normalization(self, cosmo, a, *, hmc=None):
        """Returns the normalization of all matter overdensity
        profiles, which is simply the comoving matter density.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            a (:obj:`float`): scale factor.
            hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo
                model calculator object.

        Returns:
            :obj:`float`: normalization factor of this profile.
        """
        return const.RHO_CRITICAL * cosmo["Omega_m"] * cosmo["h"]**2


class HaloProfilePressure(HaloProfile):
    """Base for pressure halo profiles."""


class HaloProfileCIB(HaloProfile):
    """Base for CIB halo profiles."""



from ..halo_model_base import Concentration, get_delta_c
from .bhattacharya13 import *
from .constant import *
from .diemer15 import *
from .duffy08 import *
from .ishiyama21 import *
from .klypin11 import *
from .prada12 import *



import numpy as np
import pyccl as ccl
import pytest

BBKS_TOLERANCE = 1.0e-5


@pytest.mark.parametrize(
    'model,w0,wa',
    [(1, -1.0, 0.0),
     (2, -0.9, 0.0),
     (3, -0.9, 0.1)])
def test_bbks(model, w0, wa):
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=0,
        m_nu=0.0,
        w0=w0,
        wa=wa,
        T_CMB=2.7,
        mass_split='normal',
        Omega_g=0,
        Omega_k=0,
        transfer_function='bbks',
        matter_power_spectrum='linear')

    data = np.loadtxt('./benchmarks/data/model%d_pk.txt' % model)

    k = data[:, 0] * cosmo['h']
    for i in range(6):
        a = 1.0 / (1.0 + i)
        pk = data[:, i+1] / (cosmo['h']**3)
        pk_ccl = ccl.linear_matter_power(cosmo, k, a)
        err = np.abs(pk_ccl/pk - 1)
        assert np.allclose(err, 0, rtol=0, atol=BBKS_TOLERANCE)



import numpy as np
import pyccl as ccl
from scipy.integrate import simps


def test_iswcl():
    # Cosmology
    Ob = 0.05
    Oc = 0.25
    h = 0.7
    COSMO = ccl.Cosmology(
        Omega_b=Ob,
        Omega_c=Oc,
        h=h,
        n_s=0.96,
        sigma8=0.8,
        transfer_function='bbks')

    # CCL calculation
    ls = np.arange(2, 100)
    zs = np.linspace(0, 0.6, 256)
    nz = np.exp(-0.5*((zs-0.3)/0.05)**2)
    bz = np.ones_like(zs)
    tr_n = ccl.NumberCountsTracer(COSMO, has_rsd=False,
                                  dndz=(zs, nz), bias=(zs, bz))
    tr_i = ccl.ISWTracer(COSMO)
    cl = ccl.angular_cl(COSMO, tr_n, tr_i, ls)

    # Benchmark from Eq. 6 in 1710.03238
    pz = nz / simps(nz, x=zs)
    H0 = h / ccl.physical_constants.CLIGHT_HMPC
    # Prefactor
    prefac = 3*COSMO['T_CMB']*(Oc+Ob)*H0**3/(ls+0.5)**2
    # H(z)/H0
    ez = ccl.h_over_h0(COSMO, 1./(1+zs))
    # Linear growth and derivative
    dz = ccl.growth_factor(COSMO, 1./(1+zs))
    gz = np.gradient(dz*(1+zs), zs[1]-zs[0])/dz
    # Comoving distance
    chi = ccl.comoving_radial_distance(COSMO, 1/(1+zs))
    # P(k)
    pks = np.array([ccl.nonlin_matter_power(COSMO, (ls+0.5)/(c+1E-6), 1./(1+z))
                    for c, z in zip(chi, zs)]).T
    # Limber integral
    cl_int = pks[:, :]*(pz*ez*gz)[None, :]
    clbb = simps(cl_int, x=zs)
    clbb *= prefac

    assert np.all(np.fabs(cl/clbb-1) < 1E-3)



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.CosmologyVanillaLCDM(transfer_function='bbks',
                                 matter_power_spectrum='linear')
M200 = ccl.halos.MassDef200m
HMF = ccl.halos.MassFuncTinker10(mass_def=M200)
HBF = ccl.halos.HaloBiasTinker10(mass_def=M200)
CON = ccl.halos.ConcentrationDuffy08(mass_def=M200)
P1 = ccl.halos.HaloProfileNFW(mass_def=M200, concentration=CON,
                              fourier_analytic=True)
P2 = ccl.halos.HaloProfileHOD(mass_def=M200, concentration=CON)
P3 = ccl.halos.HaloProfilePressureGNFW(mass_def=M200)
P4 = P1
PKC = ccl.halos.Profile2pt()
PKCH = ccl.halos.Profile2ptHOD()
KK = np.geomspace(1E-3, 10, 32)


def smoke_assert_tkk1h_real(func):
    sizes = [(0, 0),
             (2, 0),
             (0, 2),
             (2, 3),
             (1, 3),
             (3, 1)]
    shapes = [(),
              (2,),
              (2, 2,),
              (2, 3, 3),
              (1, 3, 3),
              (3, 1, 1)]
    for (sa, sk), sh in zip(sizes, shapes):
        if sk == 0:
            k = 0.1
        else:
            k = np.logspace(-2., 0., sk)
        if sa == 0:
            a = 1.
        else:
            a = np.linspace(0.5, 1., sa)
        p = func(k, a)
        assert np.shape(p) == sh
        assert np.all(np.isfinite(p))


@pytest.mark.parametrize(
    "p1,p2,cv12,p3,p4,cv34",
    [(P1, None, None, None, None, None),
     (P1, None, None, None, None, None),
     (P1, P2, None, None, None, None),
     (P1, None, None, P3, None, None),
     (P1, None, None, None, P4, None),
     (P1, P2, None, P3, P4, None),
     (P2, P2, PKCH, P2, P2, None),
     (P1, P2, PKC, P3, P4, PKC)])
def test_tkk1h_smoke(p1, p2, cv12, p3, p4, cv34):
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200, nM=2)

    def f(k, a):
        return ccl.halos.halomod_trispectrum_1h(
            COSMO, hmc, k, a,
            prof=p1, prof2=p2, prof12_2pt=cv12,
            prof3=p3, prof4=p4, prof34_2pt=cv34)
    smoke_assert_tkk1h_real(f)


def test_tkk1h_tk3d():
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200)
    k_arr = KK
    a_arr = np.array([0.1, 0.4, 0.7, 1.0])
    tkk_arr = ccl.halos.halomod_trispectrum_1h(
        COSMO, hmc, k_arr, a_arr,
        prof=P1, prof2=P2, prof12_2pt=PKC,
        prof3=P3, prof4=P4, prof34_2pt=PKC)

    # Input sampling
    tk3d = ccl.halos.halomod_Tk3D_1h(
        COSMO, hmc,
        prof=P1, prof2=P2, prof12_2pt=PKC,
        prof3=P3, prof4=P4, prof34_2pt=PKC,
        lk_arr=np.log(k_arr), a_arr=a_arr, use_log=True)

    tkk_arr_2 = tk3d(k_arr, a_arr)
    assert np.allclose(tkk_arr, tkk_arr_2, atol=0, rtol=1e-4)

    # Standard sampling
    tk3d = ccl.halos.halomod_Tk3D_1h(
        COSMO, hmc,
        prof=P1, prof2=P2, prof12_2pt=PKC,
        prof3=P3, prof4=P4, prof34_2pt=PKC,
        lk_arr=np.log(k_arr), use_log=True)

    tkk_arr_2 = tk3d(k_arr, a_arr)
    assert np.allclose(tkk_arr, tkk_arr_2, atol=0, rtol=1e-4)


def test_tkk1h_warns():
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200)
    a_arr = np.array([0.1, 0.4, 0.7, 1.0])

    # Negative profile in logspace
    Pneg = ccl.halos.HaloProfilePressureGNFW(mass_def=M200, P0=-1)
    with pytest.warns(ccl.CCLWarning):
        ccl.halos.halomod_Tk3D_1h(
            COSMO, hmc, P3, prof2=Pneg, prof3=P3, prof4=P3,
            lk_arr=np.log(KK), a_arr=a_arr, use_log=True)



import numpy as np
import pytest
import itertools
import pyccl as ccl


COSMO = ccl.CosmologyVanillaLCDM(transfer_function='bbks',
                                 matter_power_spectrum='linear')
COSMO.compute_nonlin_power()
M200 = ccl.halos.MassDef200m
HMF = ccl.halos.MassFuncTinker10(mass_def=M200)
HBF = ccl.halos.HaloBiasTinker10(mass_def=M200)
HMC = ccl.halos.HMCalculator(
    mass_function=HMF, halo_bias=HBF, mass_def=M200, nM=2)
CON = ccl.halos.ConcentrationDuffy08(mass_def=M200)

NFW = ccl.halos.HaloProfileNFW(mass_def=M200, concentration=CON,
                               fourier_analytic=True)
HOD = ccl.halos.HaloProfileHOD(mass_def=M200, concentration=CON)
HOD_nogc = ccl.halos.HaloProfileHOD(mass_def=M200, concentration=CON)
HOD_nogc.is_number_counts = False
GNFW = ccl.halos.HaloProfilePressureGNFW(mass_def=M200)

PKC = ccl.halos.Profile2pt()
PKCH = ccl.halos.Profile2ptHOD()
KK = np.geomspace(1E-3, 10, 32)
AA = np.linspace(0.3, 1.0, 4)


@pytest.mark.parametrize(
    "p1,p2,cv12,p3,p4,cv34,pk",
    [(NFW, None, None, None, None, None, None),
     (HOD, None, None, None, None, None, None),
     (NFW, HOD, None, None, None, None, None),
     (NFW, None, None, GNFW, None, None, None),
     (NFW, None, None, None, NFW, None, None),
     (NFW, HOD, None, GNFW, NFW, None, None),
     (HOD, HOD, PKCH, HOD, HOD, None, None),
     (NFW, HOD, PKC, GNFW, NFW, PKC, None),
     (NFW, None, None, None, None, None, "linear"),
     (NFW, None, None, None, None, None, "nonlinear"),
     (NFW, None, None, None, None, None, COSMO.get_nonlin_power())])
def test_tkkssc_smoke(p1, p2, cv12, p3, p4, cv34, pk):
    tkk = ccl.halos.halomod_Tk3D_SSC(
        COSMO, HMC,
        prof=p1, prof2=p2, prof12_2pt=cv12,
        prof3=p3, prof4=p4, prof34_2pt=cv34,
        p_of_k_a=pk, lk_arr=np.log(KK), a_arr=AA)

    assert (np.isfinite(tkk(0.1, 0.5))).all()


def test_tkkssc_linear_bias_smoke():
    tkkl = ccl.halos.halomod_Tk3D_SSC_linear_bias(
        COSMO, HMC, prof=NFW, p_of_k_a="linear")
    *_, tkkl_arrs = tkkl.get_spline_arrays()
    assert all([(np.isfinite(tk)).all() for tk in tkkl_arrs])

    tkknl = ccl.halos.halomod_Tk3D_SSC_linear_bias(
        COSMO, HMC, prof=NFW, p_of_k_a="nonlinear")
    *_, tkknl_arrs = tkknl.get_spline_arrays()
    assert all([(np.isfinite(tk)).all() for tk in tkknl_arrs])

    tkk_pk = ccl.halos.halomod_Tk3D_SSC_linear_bias(
        COSMO, HMC, prof=NFW, p_of_k_a=COSMO.get_nonlin_power())
    *_, tkk_pk_arrs = tkk_pk.get_spline_arrays()
    assert all([(np.isfinite(tk)).all() for tk in tkk_pk_arrs])


@pytest.mark.parametrize(
    "isNC1,isNC2,isNC3,isNC4",
    itertools.product([True, False], repeat=4))
def test_tkkssc_linear_bias(isNC1, isNC2, isNC3, isNC4):
    bias1, bias2, bias3, bias4 = 2, 3, 4, 5

    # Test that if we remove the biases 12 will be the same as 34.
    tkkl = ccl.halos.halomod_Tk3D_SSC_linear_bias(
        COSMO, HMC, prof=NFW,
        bias1=bias1, bias2=bias2, bias3=bias3, bias4=bias4,
        is_number_counts1=False, is_number_counts2=False,
        is_number_counts3=False, is_number_counts4=False,
        lk_arr=np.log(KK), a_arr=AA)

    *_, (tkkl_12, tkkl_34) = tkkl.get_spline_arrays()
    tkkl_12 /= (bias1 * bias2)
    tkkl_34 /= (bias3 * bias4)
    assert np.allclose(tkkl_12, tkkl_34, atol=0, rtol=1e-12)

    # Test with the full T(k1,k2,a) for an NFW profile with bias ~1.
    tkk = ccl.halos.halomod_Tk3D_SSC(
        COSMO, HMC, prof=NFW, lk_arr=np.log(KK), a_arr=AA)
    *_, (tkk_12, tkk_34) = tkk.get_spline_arrays()
    assert np.allclose(tkkl_12, tkk_12, atol=0, rtol=5e-3)
    assert np.allclose(tkkl_34, tkk_34, atol=0, rtol=5e-3)

    # Test with clustering profile.
    tkkl_nc = ccl.halos.halomod_Tk3D_SSC_linear_bias(
        COSMO, HMC, prof=NFW,
        bias1=bias1, bias2=bias2, bias3=bias3, bias4=bias4,
        is_number_counts1=isNC1, is_number_counts2=isNC2,
        is_number_counts3=isNC3, is_number_counts4=isNC4,
        lk_arr=np.log(KK), a_arr=AA)
    *_, (tkkl_nc_12, tkkl_nc_34) = tkkl_nc.get_spline_arrays()
    tkkl_nc_12 /= (bias1 * bias2)
    tkkl_nc_34 /= (bias3 * bias4)

    # recover the factors
    factor12 = isNC1*bias1 + isNC2*bias2
    factor34 = isNC3*bias3 + isNC4*bias4

    # calculate what the Tkk's would be with the counterterms.
    tkkl_ct_12 = (tkkl_12 - tkkl_nc_12) / factor12 if factor12 else None
    tkkl_ct_34 = (tkkl_34 - tkkl_nc_34) / factor34 if factor34 else None

    if factor12*factor34:
        assert np.allclose(tkkl_ct_12, tkkl_ct_34, atol=0, rtol=1e-5)


def test_tkkssc_warns():
    """Test that it warns if the profile is negative and use_log is True."""
    Pneg = ccl.halos.HaloProfilePressureGNFW(P0=-1, mass_def=HMC.mass_def)
    with pytest.warns(ccl.CCLWarning):
        ccl.halos.halomod_Tk3D_SSC(
            COSMO, HMC, prof=GNFW, prof2=Pneg,
            lk_arr=np.log(KK), a_arr=AA, use_log=True)


def test_tkkssc_linear_bias_raises():
    """Test that it raises if the profile is not NFW."""
    with pytest.raises(TypeError):
        ccl.halos.halomod_Tk3D_SSC_linear_bias(COSMO, HMC, prof=GNFW)


def get_ssc_counterterm_gc(k, a, hmc, prof1, prof2, prof12_2pt):
    P_12 = b1 = b2 = np.zeros_like(k)
    if prof1.is_number_counts or prof2.is_number_counts:
        norm1 = prof1.get_normalization(COSMO, a, hmc=hmc)
        norm2 = prof2.get_normalization(COSMO, a, hmc=hmc)

        i11_1 = hmc.I_1_1(COSMO, k, a, prof1)/norm1
        i11_2 = hmc.I_1_1(COSMO, k, a, prof2)/norm2
        i02_12 = hmc.I_0_2(COSMO, k, a, prof1,
                           prof_2pt=prof12_2pt,
                           prof2=prof2)/(norm1*norm2)

        pk = ccl.linear_matter_power(COSMO, k, a)
        P_12 = pk * i11_1 * i11_2 + i02_12

        if prof1.is_number_counts:
            b1 = ccl.halos.halomod_bias_1pt(COSMO, hmc, k, a, prof1)
        if prof2.is_number_counts:
            b2 = ccl.halos.halomod_bias_1pt(COSMO, hmc, k, a, prof2)

    return (b1 + b2) * P_12


@pytest.mark.parametrize('kwargs', [
                         # All is_number_counts = False
                         {'prof': NFW, 'prof2': NFW,
                          'prof3': NFW, 'prof4': NFW},
                         {'prof': HOD, 'prof2': NFW,
                          'prof3': NFW, 'prof4': NFW},
                         {'prof': HOD, 'prof2': HOD,
                          'prof3': NFW, 'prof4': NFW},
                         {'prof': HOD, 'prof2': HOD,
                          'prof3': HOD, 'prof4': NFW},
                         {'prof': NFW, 'prof2': NFW,
                          'prof3': HOD, 'prof4': HOD},
                         {'prof': HOD, 'prof2': NFW,
                          'prof3': HOD, 'prof4': HOD},
                         {'prof': HOD, 'prof2': HOD,
                          'prof3': NFW, 'prof4': HOD},
                         {'prof': HOD, 'prof2': None,
                          'prof3': NFW, 'prof4': None},
                         {'prof': HOD, 'prof2': None,
                          'prof3': None, 'prof4': None},
                         {'prof': HOD, 'prof2': HOD,
                          'prof3': None, 'prof4': None},
                         # As in benchmarks/test_covariances.py
                         {'prof': NFW, 'prof2': NFW,
                          'prof3': None, 'prof4': None},
                         # Setting prof34_2pt
                         {'prof': NFW, 'prof2': NFW,
                          'prof3': None, 'prof4': None,
                          'prof34_2pt': PKC},
                         # All is_number_counts = True
                         {'prof': HOD, 'prof2': HOD,
                          'prof3': HOD, 'prof4': HOD},
                         ]
                         )
def test_tkkssc_counterterms_gc(kwargs):
    k_arr = KK
    a_arr = np.array([0.3, 0.5, 0.7, 1.0])

    # Tk's without clustering terms. Set is_number_counts=False for HOD
    # profiles
    kwargs_nogc = kwargs.copy()
    keys = list(kwargs.keys())
    for k in keys:
        v = kwargs[k]
        if isinstance(v, ccl.halos.HaloProfileHOD):
            kwargs_nogc[k] = HOD_nogc
    tkk_nogc = ccl.halos.halomod_Tk3D_SSC(COSMO, HMC,
                                          lk_arr=np.log(k_arr), a_arr=a_arr,
                                          **kwargs_nogc)
    _, _, _, tkk_nogc_arrs = tkk_nogc.get_spline_arrays()
    tk_nogc_12, tk_nogc_34 = tkk_nogc_arrs

    # Tk's with clustering terms
    tkk_gc = ccl.halos.halomod_Tk3D_SSC(COSMO, HMC,
                                        lk_arr=np.log(k_arr), a_arr=a_arr,
                                        **kwargs)
    _, _, _, tkk_gc_arrs = tkk_gc.get_spline_arrays()
    tk_gc_12, tk_gc_34 = tkk_gc_arrs

    # Update the None's to their corresponding values
    if kwargs['prof2'] is None:
        kwargs['prof2'] = kwargs['prof']
    if kwargs['prof3'] is None:
        kwargs['prof3'] = kwargs['prof']
    if kwargs['prof4'] is None:
        kwargs['prof4'] = kwargs['prof2']

    # Tk's of the clustering terms
    tkc12 = []
    tkc34 = []
    for aa in a_arr:
        tkc12.append(get_ssc_counterterm_gc(k_arr, aa, HMC, kwargs['prof'],
                                            kwargs['prof2'], PKC))
        tkc34.append(get_ssc_counterterm_gc(k_arr, aa, HMC, kwargs['prof3'],
                                            kwargs['prof4'], PKC))
    tkc12 = np.array(tkc12)
    tkc34 = np.array(tkc34)

    assert np.abs((tk_nogc_12 - tkc12) / tk_gc_12 - 1).max() < 1e-5
    assert np.abs((tk_nogc_34 - tkc34) / tk_gc_34 - 1).max() < 1e-5



import os
import numpy as np
import pyccl as ccl

# Set cosmology
cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_g=0, Omega_k=0,
                      h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                      w0=-1, wa=0, T_CMB=2.7255,
                      transfer_function='eisenstein_hu')
# Read data
dirdat = os.path.join(os.path.dirname(__file__), 'data')

# Redshifts
zs = np.array([0., 0.5, 1.])


def test_hbf_tinker10():
    d_hbf = np.loadtxt(os.path.join(dirdat, 'hbf_tinker10.txt'),
                       unpack=True)
    mf = ccl.halos.HaloBiasTinker10()
    m = d_hbf[0]
    for iz, z in enumerate(zs):
        hb_d = d_hbf[iz+1]
        hb_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(hb_h / hb_d - 1) < 1E-3)


def test_hbf_sheth01():
    d_hbf = np.loadtxt(os.path.join(dirdat, 'hbf_sheth01.txt'),
                       unpack=True)
    mf = ccl.halos.HaloBiasSheth01()
    m = d_hbf[0]
    for iz, z in enumerate(zs):
        hb_d = d_hbf[iz+1]
        hb_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(hb_h / hb_d - 1) < 1E-3)


def test_hbf_bhattacharya11():
    d_hbf = np.loadtxt(os.path.join(dirdat, 'hbf_bhattacharya11.txt'),
                       unpack=True)
    mf = ccl.halos.HaloBiasBhattacharya11()
    m = d_hbf[0]
    for iz, z in enumerate(zs):
        hb_d = d_hbf[iz+1]
        hb_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(hb_h / hb_d - 1) < 1E-3)



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
HBFS = [ccl.halos.HaloBiasSheth99,
        ccl.halos.HaloBiasSheth01,
        ccl.halos.HaloBiasTinker10,
        ccl.halos.HaloBiasBhattacharya11]
MS = [1E13, [1E12, 1E15], np.array([1E12, 1E15])]
MFOF = ccl.halos.MassDef('fof', 'matter')
MVIR = ccl.halos.MassDef('vir', 'critical')
MDFS = [MVIR, MVIR, MFOF, MVIR]


@pytest.mark.parametrize('bM_class', HBFS)
def test_bM_subclasses_smoke(bM_class):
    bM = bM_class()
    for m in MS:
        b = bM(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


@pytest.mark.parametrize('bM_pair', zip(HBFS, MDFS))
def test_bM_mdef_raises(bM_pair):
    bM_class, mdef = bM_pair
    with pytest.raises(ValueError):
        bM_class(mass_def=mdef)


def test_bM_SO_allgood():
    bM = ccl.halos.HaloBiasTinker10(mass_def=MVIR)
    for m in MS:
        b = bM(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


@pytest.mark.parametrize('name', ['Tinker10', 'Sheth99'])
def test_bM_from_string(name):
    bM_class = ccl.halos.HaloBias.from_name(name)
    bM = bM_class()
    for m in MS:
        b = bM(COSMO, m, 0.9)
        assert np.all(np.isfinite(b))
        assert np.shape(b) == np.shape(m)


def test_bM_from_string_raises():
    with pytest.raises(KeyError):
        ccl.halos.HaloBias.from_name('Tinker11')



__all__ = ("Tk3D",)

import numpy as np

from . import CCLObject, check, lib
from .pyutils import _get_spline2d_arrays, _get_spline3d_arrays


class Tk3D(CCLObject):
    """A container for \"isotropized\" connected trispectra, relevant for
    covariance matrix calculations. These are functions of 3 variables of the
    form :math:`T(k_1,k_2,a)`, where :math:`k_i` are wavenumbers,
    and :math:`a` is the scale factor. This function can be provided as
    a 3D array (one dimension per variable), or as two 2D arrays
    corresponding to functions :math:`f_i(k,a)` such that

    .. math::
        T(k_1,k_2,a) = f_1(k_1,a)\\,f_2(k_2,a)

    Typical uses for these objects will be:

    * To store perturbation theory or halo model \"isotropized\"
      connected trispectra of the form:

      .. math::
          \\bar{T}_{abcd}(k_1, k_2, a) = \\int \\frac{d\\varphi_1}{2\\pi}
          \\int \\frac{d\\varphi_2}{2\\pi}
          T_{abcd}({\\bf k_1},-{\\bf k_1},{\\bf k_2},-{\\bf k_2}),

      where :math:`{\\bf k}_i\\equiv k_i(\\cos\\varphi_i,\\sin\\varphi_i,0)`,
      and :math:`T_{abcd}({\\bf k}_a,{\\bf k}_b,{\\bf k}_c,{\\bf k}_d)` is
      the connected trispectrum of fields :math:`\\{a,b,c,d\\}`.

    * To store the kernel for super-sample covariance calculations as a
      product of the responses of power spectra to long-wavelength
      overdensity modes :math:`\\delta_L`:

    .. math::
        \\bar{T}_{abcd}(k_1,k_2,a)=
        \\frac{\\partial P_{ab}(k_1,a)}{\\partial\\delta_L}\\,
        \\frac{\\partial P_{cd}(k_2,a)}{\\partial\\delta_L}.

    These objects can then be used, analogously to
    :class:`~pyccl.pk2d.Pk2D` objects, to construct the non-Gaussian
    covariance of angular power spectra via Limber integration. See
    :py:mod:`~pyccl.covariances`.

    Args:
        a_arr (array): an array holding values of the scale factor. Note
            that the trispectrum will be extrapolated as constant on
            values of the scale factor outside those held by this array.
        lk_arr (array): an array holding values of the natural logarithm
            of the wavenumber (in units of :math:`{\\rm Mpc}^{-1}`).
        tkk_arr (array): a 3D array with shape ``[na,nk,nk]``, where ``na``
            and ``nk`` are the sizes of ``a_arr`` and ``lk_arr`` respectively.
            This array should contain the values of the trispectrum
            at the values of scale factor and wavenumber held by ``a_arr``
            and ``lk_arr``. The array can hold the values of the natural
            logarithm of the trispectrum, depending on the value of
            ``is_logt``. If ``tkk_arr`` is ``None``, then it is assumed that
            the trispectrum can be factorized as described above, and
            the two functions :math:`f_i(k_i,a)` are described by
            ``pk1_arr`` and ``pk2_arr``. You are responsible for making sure
            all these arrays are sufficiently well sampled (i.e. the
            resolution of ``a_arr`` and ``lk_arr`` is high enough to sample
            the main features in the trispectrum). For reference, CCL
            will use bicubic interpolation to evaluate the trispectrum
            in the 2D space of wavenumbers :math:`(k_1,k_2)` at a fixed
            scale factor, and will use linear interpolation in the
            scale factor dimension.
        pk1_arr (array): a 2D array with shape ``[na,nk]`` describing the
            first function :math:`f_1(k,a)` that makes up a factorizable
            trispectrum :math:`T(k_1,k_2,a)=f_1(k_1,a)f_2(k_2,a)`.
            ``pk1_arr`` and ``pk2_arr`` are ignored if ``tkk_arr`` is not
            ``None``.
        pk2_arr (array): a 2D array with shape ``[na,nk]`` describing the
            second factor :math:`f_2(k,a)` for a factorizable trispectrum.
        is_logt (:obj:`bool`): if True, ``tkk_arr``/``pk1_arr``/``pk2_arr``
            hold the natural logarithm of the trispectrum (or its factors).
            Otherwise, the true values of the corresponding quantities are
            expected. Note that arrays will be interpolated in log space
            if ``is_logt`` is set to ``True``.
        extrap_order_lok (:obj:`int`): extrapolation order to be used on
            k-values below the minimum of the splines (use 0 or 1). Note
            that the extrapolation will be done in either
            :math:`\\log(T(k_1,k_2,a))` or :math:`T(k_1,k_2,a)`,
            depending on the value of ``is_logt``.
        extrap_order_hik (:obj:`int`): same as ``extrap_order_lok`` for
            k-values above the maximum of the splines.

    .. automethod:: __call__
    """
    from ._core.repr_ import build_string_Tk3D as __repr__

    def __init__(self, *, a_arr, lk_arr, tkk_arr=None,
                 pk1_arr=None, pk2_arr=None, is_logt=True,
                 extrap_order_lok=1, extrap_order_hik=1):
        na = len(a_arr)
        nk = len(lk_arr)

        if not (np.diff(a_arr) > 0).all():
            raise ValueError("`a_arr` must be strictly increasing")

        if not np.all(lk_arr[1:]-lk_arr[:-1] > 0):
            raise ValueError("`lk_arr` must be strictly increasing")

        if ((extrap_order_hik not in (0, 1)) or
                (extrap_order_lok not in (0, 1))):
            raise ValueError("Only constant or linear extrapolation in "
                             "log(k) is possible (`extrap_order_hik` or "
                             "`extrap_order_lok` must be 0 or 1).")
        status = 0

        if tkk_arr is None:
            if pk2_arr is None:
                pk2_arr = pk1_arr
            if (pk1_arr is None) or (pk2_arr is None):
                raise ValueError("If trispectrum is factorizable "
                                 "you must provide the two factors")
            if (pk1_arr.shape != (na, nk)) or (pk2_arr.shape != (na, nk)):
                raise ValueError("Input trispectrum factor "
                                 "shapes are wrong")

            self.tsp, status = lib.tk3d_new_factorizable(lk_arr, a_arr,
                                                         pk1_arr.flatten(),
                                                         pk2_arr.flatten(),
                                                         int(extrap_order_lok),
                                                         int(extrap_order_lok),
                                                         int(is_logt), status)
        else:
            if tkk_arr.shape != (na, nk, nk):
                raise ValueError("Input trispectrum shape is wrong")

            self.tsp, status = lib.tk3d_new_from_arrays(lk_arr, a_arr,
                                                        tkk_arr.flatten(),
                                                        int(extrap_order_lok),
                                                        int(extrap_order_lok),
                                                        int(is_logt), status)
        check(status)

    def __eq__(self, other):
        # Check object id.
        if self is other:
            return True
        # Check the object class.
        if type(self) is not type(other):
            return False
        # If the objects contain no data, return early.
        if not (self or other):
            return True
        # If one is factorizable and the other one is not, return early.
        if self.tsp.is_product ^ other.tsp.is_product:
            return False
        # Check extrapolation orders.
        if not (self.extrap_order_lok == other.extrap_order_lok
                and self.extrap_order_hik == other.extrap_order_hik):
            return False
        # Check the individual splines.
        a1, lk11, lk12, tk1 = self.get_spline_arrays()
        a2, lk21, lk22, tk2 = other.get_spline_arrays()
        return ((a1 == a2).all()
                and (lk11 == lk21).all() and (lk21 == lk22).all()
                and np.array_equal(tk1, tk2))

    def __hash__(self):
        return hash(repr(self))

    @property
    def has_tsp(self):
        return 'tsp' in vars(self)

    @property
    def extrap_order_lok(self):
        return self.tsp.extrap_order_lok if self else None

    @property
    def extrap_order_hik(self):
        return self.tsp.extrap_order_hik if self else None

    def __call__(self, k, a):
        """Evaluate trispectrum. If ``k`` is a 1D array with size ``nk``, and
        ``a`` is a scalar, the output ``out`` will be a 2D array with shape
        ``[nk,nk]`` holding ``out[i,j] = T(k[j],k[i],a)``, where ``T`` is the
        trispectrum function held by this :class:`Tk3D` object. If ``a`` is
        an array, the shape will be ``[len(a),nk,nk]``.

        Args:
            k (:obj:`float` or `array`): wavenumber value(s) in units of
                :math:`{\\rm Mpc}^{-1}`.
            a (:obj:`float` or `array`): value(s) of the scale factor

        Returns:
            (:obj:`float` or `array`): value(s) of the trispectrum.
        """
        a_use = np.atleast_1d(a).astype(float)
        k_use = np.atleast_1d(k).astype(float)
        lk_use = np.log(k_use)

        nk = k_use.size
        out = np.zeros([len(a_use), nk, nk])

        status = 0
        for ia, aa in enumerate(a_use):
            f, status = lib.tk3d_eval_multi(self.tsp, lk_use,
                                            aa, nk*nk, status)
            check(status)
            out[ia] = f.reshape([nk, nk])

        if np.ndim(k) == 0:
            out = np.squeeze(np.squeeze(out, axis=-1), axis=-1)
        if np.ndim(a) == 0:
            out = np.squeeze(out, axis=0)
        return out

    def __del__(self):
        if hasattr(self, 'has_tsp'):
            if self.has_tsp and hasattr(self, 'tsp'):
                lib.f3d_t_free(self.tsp)

    def __bool__(self):
        return self.has_tsp

    def get_spline_arrays(self):
        """Get the spline data arrays.

        Returns:
            Tuple containing

            - a_arr (1D ``numpy.ndarray``): Array of scale factors.
            - lk_arr1, lk_arr2 (1D ``numpy.ndarray``): Arrays of
              :math:`log(k)`.
            - out (list of ``numpy.ndarray``): The trispectrum
              :math:`T(k_1, k_2, z)` or its factors
              :math:`f(k_1, z),\\,\\,f(k_2, z)`.
        """
        if not self:
            raise ValueError("Tk3D object does not have data.")

        out = []
        if self.tsp.is_product:
            a_arr, lk_arr1, pk_arr1 = _get_spline2d_arrays(self.tsp.fka_1.fka)
            _, lk_arr2, pk_arr2 = _get_spline2d_arrays(self.tsp.fka_2.fka)
            out.append(pk_arr1)
            out.append(pk_arr2)
        else:
            status = 0
            a_arr, status = lib.get_array(self.tsp.a_arr, self.tsp.na, status)
            check(status)
            lk_arr1, lk_arr2, tkka_arr = _get_spline3d_arrays(self.tsp.tkka,
                                                              self.tsp.na)
            out.append(tkka_arr)

        if self.tsp.is_log:
            # exponentiate in-place
            [np.exp(tk, out=tk) for tk in out]

        return a_arr, lk_arr1, lk_arr2, out



import pyccl
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
import astropy.cosmology as csmm
from astropy.constants import c,G
from astropy import units as u
#If you have issues running this script, contact Sukhdeep Singh (nabsukh@gmail.com)

cosmo=csmm.FlatLambdaCDM(70.,0.3,Tcmb0=2.725)
cosmo_h=cosmo.clone(H0=100)
c=c.to(u.km/u.second)

cosmo_fid=dict({'h':0.7,'Omb':0.0,'Omd':0.3,'s8':0.8,'Om':0.3,
                'As':2.12e-09,'mnu':0.,'Omk':0.,'tau':0.06,'ns':0.96})
pk_params={'non_linear':1,'kmax':30,'kmin':1.e-4,'nk':5000}

class Power_Spectra():
    def __init__(self,cosmo_params=cosmo_fid,pk_params=pk_params,cosmo=cosmo,cosmo_h=None):
        self.cosmo_params=cosmo_params
        self.pk_params=pk_params
        self.cosmo=cosmo

        if not cosmo_h:
            self.cosmo_h=cosmo.clone(H0=100)
        else:
            self.cosmo_h=cosmo_h
            
    def ccl_pk(self,z,cosmo_params=None,pk_params=None):
        if not cosmo_params:
            cosmo_params=self.cosmo_params
        if not pk_params:
            pk_params=self.pk_params

        cosmo_ccl=pyccl.Cosmology(h=cosmo_params['h'],Omega_c=cosmo_params['Omd'],
                                  Omega_b=cosmo_params['Omb'],sigma8=0.8,
                                  n_s=cosmo_params['ns'],m_nu=cosmo_params['mnu'],
                                  transfer_function='bbks')
        kh=np.logspace(np.log10(pk_params['kmin']),np.log10(pk_params['kmax']),pk_params['nk'])
        nz=len(z)
        ps=np.zeros((nz,pk_params['nk']))
        ps0=[]
        z0=9.#PS(z0) will be rescaled using growth function when CCL fails. 

        pyccl_pkf=pyccl.linear_matter_power
        if pk_params['non_linear']==1:
            pyccl_pkf=pyccl.nonlin_matter_power
        for i in np.arange(nz):
            ps[i]= pyccl_pkf(cosmo_ccl,kh,1./(1+z[i]))
        return ps*cosmo_params['h']**3,kh/cosmo_params['h'] #factors of h to get in same units as camb output

    def cl_z(self,z=[],l=np.arange(3000)+1,pk_params=None,cosmo_h=None,cosmo=None,pk_func=None):
        if not cosmo_h:
            cosmo_h=self.cosmo_h

        nz=len(z)
        nl=len(l)

        pk,kh=pk_func(z=z,pk_params=pk_params)

        cls=np.zeros((nz,nl),dtype='float32')*u.Mpc**2
        for i in np.arange(nz):
            DC_i=cosmo_h.comoving_transverse_distance(z[i]).value#because camb k in h/mpc
            lz=kh*DC_i-0.5
            DC_i=cosmo_h.comoving_transverse_distance(z[i]).value
            pk_int=interp1d(lz,pk[i]/DC_i**2,bounds_error=False,fill_value=0)
            cls[i][:]+=pk_int(l)*u.Mpc*(c/(cosmo_h.efunc(z[i])*cosmo_h.H0))
        return cls

    def kappa_cl(self,zl_min=0,zl_max=1100,n_zl=10,log_zl=False,pk_func=None,
                zs1=[1100],p_zs1=[1],zs2=[1100],p_zs2=[1],
                pk_params=None,cosmo_h=None,l=np.arange(3001)):
        if not cosmo_h:
            cosmo_h=self.cosmo_h

        if log_zl:#bins for z_lens. 
            zl=np.logspace(np.log10(max(zl_min,1.e-4)),np.log10(zl_max),n_zl)
        else:
            zl=np.linspace(zl_min,zl_max,n_zl)

        ds1=cosmo_h.comoving_transverse_distance(zs1)
        ds2=cosmo_h.comoving_transverse_distance(zs2) 
        dl=cosmo_h.comoving_transverse_distance(zl)
        prefac=1.5*(cosmo_h.H0/c)**2*cosmo_h.Om0
        sigma_c1=prefac*(1+zl)*dl*(1-np.multiply.outer(1./ds1,dl))
        sigma_c2=prefac*(1+zl)*dl*(1-np.multiply.outer(1./ds2,dl))
        clz=self.cl_z(z=zl,l=l,cosmo_h=cosmo_h,pk_params=pk_params,pk_func=pk_func)

        dzl=np.gradient(zl)
        dzs1=np.gradient(zs1) if len(zs1)>1 else 1
        dzs2=np.gradient(zs2) if len(zs2)>1 else 1

        cl_zs_12=np.einsum('ji,ki,il',sigma_c2,sigma_c1*dzl,clz)#integrate over zl..
        cl=np.dot(p_zs2*dzs2,np.dot(p_zs1*dzs1,cl_zs_12))
        cl/=np.sum(p_zs2*dzs2)*np.sum(p_zs1*dzs1)
        f=l*(l+1.)/(l+0.5)**2 #correction from Kilbinger+ 2017
        cl*=f**2
        return l,cl


if __name__ == "__main__":
    PS=Power_Spectra()
    l,cl2=PS.kappa_cl(n_zl=10000,log_zl=True,zl_min=1.e-4,zl_max=1100,pk_func=PS.ccl_pk)
    
    plt.figure()
    plt.plot(l,cl2)
    plt.loglog()
    plt.xlabel('$\\ell$',fontsize=16)
    plt.ylabel('$C^{\\kappa\\kappa}_\\ell$',fontsize=16)
    plt.show()
    
    np.savetxt('benchmark_cc.txt',np.transpose([l,cl2]),fmt='%d %lE')



__all__ = ("Profile2pt", "Profile2ptHOD", "Profile2ptCIB",)

from .. import CCLAutoRepr
from . import HaloProfileHOD, HaloProfileCIBShang12


class Profile2pt(CCLAutoRepr):
    """ This class implements the 1-halo 2-point correlator between
    two halo profiles.

    .. math::
        \\langle u_1(k) u_2(k) \\rangle.

    In the simplest case the second-order cumulant is just the product
    of the individual Fourier-space profiles. More complicated cases
    are implemented via the parameters of this class.

    Args:
        r_corr (:obj:`float`):
            Tuning knob for the 1-halo 2-point correlation.
            Scale the correlation by :math:`(1+\\rho_{u_1, u_2})`.
            This is useful when the individual 1-halo terms
            are not fully correlated. Example usecases can be found
            in ``arXiv:1909.09102`` and ``arXiv:2102.07701``.
            Defaults to ``r_corr=0``, returning simply the product
            of the fourier profiles.
    """
    __repr_attrs__ = __eq_attrs__ = ("r_corr",)

    def __init__(self, *, r_corr=0.):
        self.r_corr = r_corr

    def update_parameters(self, *, r_corr=None):
        """ Update any of the parameters associated with this 1-halo
        2-point correlator. Any parameter set to ``None`` won't be updated.
        """
        if r_corr is not None:
            self.r_corr = r_corr

    def fourier_2pt(self, cosmo, k, M, a, prof, *, prof2=None):
        """ Return the Fourier-space two-point moment between
        two profiles.

        .. math::
           (1+\\rho_{u_1,u_2})\\langle u_1(k)\\rangle\\langle u_2(k) \\rangle

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                a Cosmology object.
            k (:obj:`float` or `array`):
                comoving wavenumber in Mpc^-1.
            M (:obj:`float` or `array`):
                halo mass in units of M_sun.
            a (:obj:`float`):
                scale factor.
            prof (:class:`~pyccl.halos.profiles.HaloProfile`):
                halo profile for which the second-order moment
                is desired.
            prof2 (:class:`~pyccl.halos.profiles.HaloProfile`):
                second halo profile for which the second-order moment
                is desired. If ``None``, the assumption is that you want
                an auto-correlation, and `prof` will be used as `prof2`.

        Returns:
            (:obj:`float` or `array`): second-order Fourier-space
            moment. The shape of the output will be `(N_M, N_k)`
            where `N_k` and `N_m` are the sizes of `k` and `M`
            respectively. If `k` or `M` are scalars, the
            corresponding dimension will be squeezed out on output.
        """
        if prof2 is None:
            prof2 = prof

        uk1 = prof.fourier(cosmo, k, M, a)

        if prof == prof2:
            uk2 = uk1
        else:
            uk2 = prof2.fourier(cosmo, k, M, a)

        return uk1 * uk2 * (1 + self.r_corr)


class Profile2ptHOD(Profile2pt):
    """ This class implements the Fourier-space 1-halo 2-point
    correlator for the HOD profile.

    .. math::
       \\langle n_g^2(k)|M,a\\rangle = \\bar{N}_c(M,a)
       \\left[2f_c(a)\\bar{N}_s(M,a) u_{\\rm sat}(r|M,a)+
       (\\bar{N}_s(M,a) u_{\\rm sat}(r|M,a))^2\\right],

    where all quantities are described in the documentation of
    :class:`~pyccl.halos.profiles.hod.HaloProfileHOD`.
    """

    def fourier_2pt(self, cosmo, k, M, a, prof, *, prof2=None):
        """ Returns the Fourier-space two-point moment for the HOD
        profile.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber in Mpc^-1.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.hod.HaloProfileHOD`):
                halo profile for which the second-order moment
                is desired.
            prof2 (:class:`~pyccl.halos.profiles.hod.HaloProfileHOD` or :obj:`None`):
                second halo profile for which the second-order moment
                is desired. If ``None``, the assumption is that you want
                an auto-correlation. Note that only auto-correlations
                are allowed in this case.

        Returns:
            (:obj:`float` or `array`): second-order Fourier-space
            moment. The shape of the output will be `(N_M, N_k)`
            where `N_k` and `N_m` are the sizes of `k` and `M`
            respectively. If `k` or `M` are scalars, the
            corresponding dimension will be squeezed out on output.
        """ # noqa
        if prof2 is None:
            prof2 = prof

        if prof != prof2:
            raise ValueError("prof and prof2 must be equivalent")
        HOD = HaloProfileHOD
        if not (isinstance(prof, HOD) and isinstance(prof2, HOD)):
            raise TypeError("prof and prof2 must be HaloProfileHOD")

        return prof._fourier_variance(cosmo, k, M, a)


class Profile2ptCIB(Profile2pt):
    """ This class implements the Fourier-space 1-halo 2-point
    correlator for the CIB profile. It follows closely the
    implementation of the equivalent HOD quantity
    (see :class:`~pyccl.halos.profiles_2pt.Profile2ptHOD`
    and Eq. 15 of `McCarthy & Madhavacheril
    <https://arxiv.org/abs/2010.16405>`_).
    """

    def fourier_2pt(self, cosmo, k, M, a, prof, *, prof2=None):
        """ Returns the Fourier-space two-point moment for the CIB
        profile.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            k (:obj:`float` or `array`): comoving wavenumber in Mpc^-1.
            M (:obj:`float` or `array`): halo mass in units of M_sun.
            a (:obj:`float`): scale factor.
            prof (:class:`~pyccl.halos.profiles.cib_shang12.HaloProfileCIBShang12`):
                halo profile for which the second-order moment
                is desired.
            prof2 (:class:`~pyccl.halos.profiles.cib_shang12.HaloProfileCIBShang12`):
                second halo profile for which the second-order moment
                is desired. If ``None``, the assumption is that you want
                an auto-correlation. Note that only auto-correlations
                are allowed in this case.

        Returns:
            (:obj:`float` or `array`): second-order Fourier-space
            moment. The shape of the output will be `(N_M, N_k)`
            where `N_k` and `N_m` are the sizes of `k` and `M`
            respectively. If `k` or `M` are scalars, the
            corresponding dimension will be squeezed out on output.
        """ # noqa
        if prof2 is None:
            prof2 = prof

        Shang12 = HaloProfileCIBShang12
        if not (isinstance(prof, Shang12) and isinstance(prof2, Shang12)):
            raise TypeError("prof and prof2 must be HaloProfileCIB")

        return prof._fourier_variance(cosmo, k, M, a, nu_other=prof2.nu)



__all__ = ("HaloProfileHernquist",)

import numpy as np
from scipy.special import sici

from . import HaloProfileMatter


class HaloProfileHernquist(HaloProfileMatter):
    """ `Hernquist 1990
    <https://ui.adsabs.harvard.edu/abs/1990ApJ...356..359H/abstract>`_
    profile.

    .. math::
       \\rho(r) = \\frac{\\rho_0}
       {\\frac{r}{r_s}\\left(1+\\frac{r}{r_s}\\right)^3}

    where :math:`r_s` is related to the comoving spherical overdensity
    halo radius :math:`r_\\Delta(M)` through the concentration
    parameter :math:`c(M)` as

    .. math::
       r_\\Delta(M) = c(M)\\,r_s

    and the normalization :math:`\\rho_0` is

    .. math::
       \\rho_0 = \\frac{M}{2\\pi\\,r_s^3}\\left(\\frac{1+c}{c}\\right)^2

    By default, this profile is truncated at :math:`r = r_\\Delta(M)`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        concentration (:class:`~pyccl.halos.halo_model_base.Concentration`):
            concentration-mass relation to use with this profile.
        fourier_analytic (:obj:`bool`): set to ``True`` if you want to compute
            the Fourier profile analytically (and not through FFTLog).
        projected_analytic (:obj:`bool`): set to ``True`` if you want to
            compute the 2D projected profile analytically (and not
            through FFTLog).
        cumul2d_analytic (:obj:`bool`): set to ``True`` if you want to
            compute the 2D cumulative surface density analytically
            (and not through FFTLog).
        truncated (:obj:`bool`): set to ``True`` if the profile should be
            truncated at :math:`r = r_\\Delta`.
    """
    __repr_attrs__ = __eq_attrs__ = (
        "fourier_analytic", "projected_analytic", "cumul2d_analytic",
        "truncated", "mass_def", "concentration", "precision_fftlog",)

    def __init__(self, *, mass_def, concentration,
                 truncated=True,
                 fourier_analytic=False,
                 projected_analytic=False,
                 cumul2d_analytic=False):
        self.truncated = truncated
        self.fourier_analytic = fourier_analytic
        self.projected_analytic = projected_analytic
        self.cumul2d_analytic = cumul2d_analytic
        if fourier_analytic:
            self._fourier = self._fourier_analytic
        if projected_analytic:
            if truncated:
                raise ValueError("Analytic projected profile not supported "
                                 "for truncated Hernquist. Set `truncated` or "
                                 "`projected_analytic` to `False`.")
            self._projected = self._projected_analytic
        if cumul2d_analytic:
            if truncated:
                raise ValueError("Analytic cumuative 2d profile not supported "
                                 "for truncated Hernquist. Set `truncated` or "
                                 "`cumul2d_analytic` to `False`.")
            self._cumul2d = self._cumul2d_analytic
        super().__init__(mass_def=mass_def, concentration=concentration)
        self.update_precision_fftlog(padding_hi_fftlog=1E2,
                                     padding_lo_fftlog=1E-4,
                                     n_per_decade=1000,
                                     plaw_fourier=-2.)

    def _norm(self, M, Rs, c):
        # Hernquist normalization from mass, radius and concentration
        return M / (2 * np.pi * Rs**3 * (c / (1 + c))**2)

    def _real(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        norm = self._norm(M_use, R_s, c_M)

        x = r_use[None, :] / R_s[:, None]
        prof = norm[:, None] / (x * (1 + x)**3)
        if self.truncated:
            prof[r_use[None, :] > R_M[:, None]] = 0

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fx_projected(self, x):

        def f1(xx):
            x2m1 = xx * xx - 1
            sqx2m1 = np.sqrt(-x2m1)
            return (-3 / 2 / x2m1**2
                    + (x2m1+3) * np.arcsinh(sqx2m1 / xx) / 2 / (-x2m1)**2.5)

        def f2(xx):
            x2m1 = xx * xx - 1
            sqx2m1 = np.sqrt(x2m1)
            return (-3 / 2 / x2m1**2
                    + (x2m1+3) * np.arcsin(sqx2m1 / xx) / 2 / x2m1**2.5)

        xf = x.flatten()
        return np.piecewise(xf,
                            [xf < 1, xf > 1],
                            [f1, f2, 2./15.]).reshape(x.shape)

    def _projected_analytic(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        x = r_use[None, :] / R_s[:, None]
        prof = self._fx_projected(x)
        norm = 2 * R_s * self._norm(M_use, R_s, c_M)
        prof = prof[:, :] * norm[:, None]

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fx_cumul2d(self, x):

        def f1(xx):
            x2m1 = xx * xx - 1
            sqx2m1 = np.sqrt(-x2m1)
            return (1 + 1 / x2m1
                    + (x2m1 + 1) * np.arcsinh(sqx2m1 / xx) / (-x2m1)**1.5)

        def f2(xx):
            x2m1 = xx * xx - 1
            sqx2m1 = np.sqrt(x2m1)
            return (1 + 1 / x2m1
                    - (x2m1 + 1) * np.arcsin(sqx2m1 / xx) / x2m1**1.5)

        xf = x.flatten()
        f = np.piecewise(xf,
                         [xf < 1, xf > 1],
                         [f1, f2, 1./3.]).reshape(x.shape)

        return f / x**2

    def _cumul2d_analytic(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        x = r_use[None, :] / R_s[:, None]
        prof = self._fx_cumul2d(x)
        norm = 2 * R_s * self._norm(M_use, R_s, c_M)
        prof = prof * norm[:, None]

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fourier_analytic(self, cosmo, k, M, a):
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        # Comoving virial radius
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M

        x = k_use[None, :] * R_s[:, None]
        Si2, Ci2 = sici(x)
        P1 = M / ((c_M / (c_M + 1))**2 / 2)
        c_Mp1 = c_M[:, None] + 1
        if self.truncated:
            Si1, Ci1 = sici(c_Mp1 * x)
            P2 = x * np.sin(x) * (Ci1 - Ci2) - x * np.cos(x) * (Si1 - Si2)
            P3 = (-1 + np.sin(c_M[:, None] * x) / (c_Mp1**2 * x)
                  + c_Mp1 * np.cos(c_M[:, None] * x) / (c_Mp1**2))
            prof = P1[:, None] * (P2 - P3) / 2
        else:
            P2 = (-x * (2 * np.sin(x) * Ci2 + np.pi * np.cos(x))
                  + 2 * x * np.cos(x) * Si2 + 2) / 4
            prof = P1[:, None] * P2

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof



import numpy as np
import MiraTitanHMFemulator


emu = MiraTitanHMFemulator.Emulator()

hmcosmo = {'Ommh2': 0.147,
           'Ombh2': 0.02205,
           'Omnuh2': 0.005,
           'n_s': 0.96,
           'h': 0.7,
           'sigma_8': 0.81,
           'w_0': -1.05,
           'w_a': 0.1}

Ms = np.geomspace(1E13, 1E15, 32)
zs = np.array([0.0, 1.0])
mfp = emu.predict(hmcosmo, zs, Ms, get_errors=False)[0]
np.savez("../mf_bocquet20mf.py", m=Ms, z=zs, mf=mfp,
         **hmcosmo)




__all__ = ("MassFuncBocquet20",)

import numpy as np

from . import MassFunc


class MassFuncBocquet20(MassFunc):
    """Implements the mass function emulator of `Bocquet et al. 2020
    <https://arxiv.org/abs/2003.12116>`_. This parametrization is
    only valid for '200c' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = 'Bocquet20'

    def __init__(self, *,
                 mass_def="200c",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)
        import MiraTitanHMFemulator
        self.emu = MiraTitanHMFemulator.Emulator()

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name != '200c'

    def __call__(self, cosmo, M, a):
        # Set up cosmology
        h = cosmo['h']
        om = cosmo['Omega_m']*h**2
        ob = cosmo['Omega_b']*h**2
        sigma8 = cosmo['sigma8']
        if np.isnan(sigma8):
            raise ValueError("sigma8 must be provided to use the MiraTitan "
                             "mass function emulator.")
        onu = cosmo.omega_x(1.0, 'neutrinos_massive')*h**2
        hmcosmo = {'Ommh2': om,
                   'Ombh2': ob,
                   'Omnuh2': onu,
                   'n_s': cosmo['n_s'],
                   'h': h,
                   'sigma_8': sigma8,
                   'w_0': cosmo['w0'],
                   'w_a': cosmo['wa']}

        # Filter out masses beyond emulator range
        M_use = np.atleast_1d(M)
        # Add h-inverse
        Mh = M_use * h
        m_hi = Mh > 1E16
        m_lo = Mh < 1E13
        m_good = ~(m_hi | m_lo)

        mfh = np.zeros_like(Mh)
        # Populate low-halo masses through extrapolation if needed
        if np.any(m_lo):
            # Evaluate slope at low masses
            m0 = 10**np.array([13.0, 13.1])
            mfp = self.emu.predict(hmcosmo, 1/a-1, m0,
                                   get_errors=False)[0].flatten()
            slope = np.log(mfp[1]/mfp[0])/np.log(m0[1]/m0[0])
            mfh[m_lo] = mfp[0]*(Mh[m_lo]/m0[0])**slope
        # Predict in good range of masses
        if np.any(m_good):
            mfp = self.emu.predict(hmcosmo, 1/a-1, Mh[m_good],
                                   get_errors=False)[0].flatten()
            mfh[m_good] = mfp
        # For masses above emulator range, n(M) will be set to zero
        # Remove h-inverse and correct for dn/dln(M) -> dn/dlog10(M)
        # log10 = 2.30258509299
        mf = mfh*h**3*2.30258509299

        if np.ndim(M) == 0:
            return mf[0]
        return mf



__all__ = ('EmulatorPk', )

from abc import abstractmethod


class EmulatorPk(object):

    @abstractmethod
    def _get_pk_at_a(self, cosmo, a):
        """Get k vector and uninterpolated power spectrum at given a.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.
            a (:obj:`float` or `array`):
                Scale factor.

        Returns:
            :tuple: k and pk arrays.
        """

    def get_pk_at_a(self, cosmo, a):
        """Get k vector and uninterpolated power spectrum at given a.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.
            a (:obj:`float` or `array`):
                Scale factor.

        Returns:
            :tuple: k and pk arrays.
        """
        return self._get_pk_at_a(cosmo, a)

    @abstractmethod
    def _get_pk2d(self, cosmo):
        """Get a 2D interpolator for the power of k and a.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.

        Returns:
            :obj:`~pyccl.pk2d.Pk2D` object.
        """

    def get_pk2d(self, cosmo):
        """Get a 2D interpolator for the power of k and a.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.

        Returns:
            :obj:`~pyccl.pk2d.Pk2D` object.
        """
        return self._get_pk2d(cosmo)



__all__ = ("BaryonsvanDaalen19",)

import numpy as np

from .. import Pk2D
from . import Baryons


class BaryonsvanDaalen19(Baryons):
    """The baryonic boost factor model of
    `van Daalen et al. 2019, <https://arxiv.org/abs/1906.00968>`_.

    The boost factor is applied multiplicatively so that
    :math:`P_{\\rm bar.}(k, a) = P_{\\rm DMO}(k, a)\\, f_{\\rm vD19}(k, a)`.

    .. note:: The model has only been tested at z=0 and is valid for
              :math:`k\\leq 1 \\,h/{\\rm Mpc}`.

    Args:
        fbar (:obj:`float`): the fraction of baryons in a halo in units
            of the ratio of :math:`\\Omega_b` to :math:`\\Omega_m`.
            Default to 0.7 which is approximately compatible with observations
            (see Fig. 16 of the paper).
        mass_def (:obj:`string`): spherical overdensity mass definition.
            Options are "500c" or "200c".

    """
    name = 'vanDaalen19'
    __repr_attrs__ = __eq_attrs__ = ("fbar", "mass_def",)

    def __init__(self, fbar=0.7, mass_def='500c'):
        self.fbar = fbar
        self.mass_def = mass_def
        if mass_def not in ["500c", "200c"]:
            raise ValueError(f"Mass definition {mass_def} not supported "
                             "for van Daalen 2019 model.")

    def boost_factor(self, cosmo, k, a):
        """The vD19 model boost factor for baryons.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
            k (:obj:`float` or `array`): Wavenumber (in :math:`{\\rm Mpc}^{-1}`).
            a (:obj:`float` or `array`): Scale factor.

        Returns:
            :obj:`float` or `array`: Correction factor to apply \
            to the power spectrum.

        """ # noqa
        a_use, k_use = map(np.atleast_1d, [a, k])
        # [k]=1/Mpc [k_use]=h/Mpc
        a_use, k_use = a_use[:, None], k_use[None, :]/cosmo['h']

        if self.mass_def == '500c':
            avD19 = 2.215
            bvD19 = 0.1276
            cvD19 = 1.309
            dvD19 = -5.99
            evD19 = -0.5107
        else:
            avD19 = 2.111
            bvD19 = 0.0038
            cvD19 = 1.371
            dvD19 = -5.816
            evD19 = -0.4005

        numf = (2**avD19 +
                2**bvD19*(cvD19*self.fbar)**(bvD19-avD19))
        expf = np.exp(dvD19*self.fbar+evD19)
        denf1 = (cvD19*self.fbar)**(bvD19-avD19)
        denf = k_use**(-avD19)+k_use**(-bvD19)*denf1
        fka = 1-numf/denf*expf

        if np.ndim(k) == 0:
            fka = np.squeeze(fka, axis=-1)
        if np.ndim(a) == 0:
            fka = np.squeeze(fka, axis=0)
        return fka

    def update_parameters(self, fbar=None, mass_def=None):
        """Update van Daalen 2019 parameters. All parameters set to
        ``None`` will be left untouched.

        Args:
            fbar (:obj:`float`): baryonic fraction in halos.
            mass_def (:obj:`string`): mass definition ("500c" or
                "200c")
        """
        if fbar is not None:
            self.fbar = fbar
        if mass_def is not None:
            self.mass_def = mass_def
            if mass_def not in ["500c", "200c"]:
                raise ValueError(f"Mass definition {mass_def} not supported "
                                 "for van Daalen 2019 model.")

    def _include_baryonic_effects(self, cosmo, pk):
        # Applies boost factor
        a_arr, lk_arr, pk_arr = pk.get_spline_arrays()
        k_arr = np.exp(lk_arr)
        fka = self.boost_factor(cosmo, k_arr, a_arr)
        pk_arr *= fka

        if pk.psp.is_log:
            np.log(pk_arr, out=pk_arr)  # in-place log

        return Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                    is_logp=pk.psp.is_log,
                    extrap_order_lok=pk.extrap_order_lok,
                    extrap_order_hik=pk.extrap_order_hik)



import os
import time
import numpy as np
import pyccl as ccl
from scipy.interpolate import interp1d
import pytest

T0 = 0.0
T0_CLS = 0.0


@pytest.fixture(scope='module', params=['fftlog', 'bessel'])
def corr_method(request):
    errfacs = {'fftlog': 0.2, 'bessel': 0.1}
    return request.param, errfacs[request.param]


@pytest.fixture(scope='module', params=['analytic', 'histo'])
def set_up(request):
    t0 = time.time()
    nztyp = request.param
    dirdat = os.path.dirname(__file__) + '/data/'
    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    ccl.gsl_params.INTEGRATION_LIMBER_EPSREL = 2.5E-5
    ccl.gsl_params.INTEGRATION_EPSREL = 2.5E-5
    cosmo = ccl.Cosmology(Omega_c=0.30, Omega_b=0.00, Omega_g=0, Omega_k=0,
                          h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                          w0=-1, wa=0, T_CMB=2.7, transfer_function='bbks',
                          matter_power_spectrum='linear')

    # Ell-dependent correction factors
    # Set up array of ells
    fl = {}
    lmax = 10000
    nls = (lmax - 400)//20+141
    ells = np.zeros(nls)
    ells[:101] = np.arange(101)
    ells[101:121] = ells[100] + (np.arange(20) + 1) * 5
    ells[121:141] = ells[120] + (np.arange(20) + 1) * 10
    ells[141:] = ells[140] + (np.arange(nls - 141) + 1) * 20
    fl['lmax'] = lmax
    fl['ells'] = ells

    # Initialize tracers
    if nztyp == 'analytic':
        # Analytic case
        zmean_1 = 1.0
        sigz_1 = 0.15
        zmean_2 = 1.5
        sigz_2 = 0.15
        z1, tmp_a1 = np.loadtxt(dirdat + "ia_amp_analytic_1.txt", unpack=True)
        z2, tmp_a2 = np.loadtxt(dirdat + "ia_amp_analytic_2.txt", unpack=True)
        pz1 = np.exp(-0.5 * ((z1 - zmean_1) / sigz_1)**2)
        pz2 = np.exp(-0.5 * ((z2 - zmean_2) / sigz_2)**2)
    elif nztyp == 'histo':
        # Histogram case
        z1, pz1 = np.loadtxt(dirdat + "bin1_histo.txt", unpack=True)[:, 1:]
        _, tmp_a1 = np.loadtxt(dirdat + "ia_amp_histo_1.txt", unpack=True)
        z2, pz2 = np.loadtxt(dirdat + "bin2_histo.txt", unpack=True)[:, 1:]
        _, tmp_a2 = np.loadtxt(dirdat + "ia_amp_histo_2.txt", unpack=True)
    else:
        raise ValueError("Wrong Nz type " + nztyp)
    bz = np.ones_like(pz1)

    # Renormalize the IA amplitude to be consistent with A_IA
    D1 = ccl.growth_factor(cosmo, 1./(1+z1))
    D2 = ccl.growth_factor(cosmo, 1./(1+z2))
    rho_m = ccl.physical_constants.RHO_CRITICAL * cosmo['Omega_m']
    a1 = - tmp_a1 * D1 / (5e-14 * rho_m)
    a2 = - tmp_a2 * D2 / (5e-14 * rho_m)

    # Initialize tracers
    trc = {}
    trc['g1'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z1, pz1),
                                       bias=(z2, bz))
    trc['g2'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z2, pz2),
                                       bias=(z2, bz))
    trc['l1'] = ccl.WeakLensingTracer(cosmo, dndz=(z1, pz1))
    trc['l2'] = ccl.WeakLensingTracer(cosmo, dndz=(z2, pz2))
    trc['i1'] = ccl.WeakLensingTracer(cosmo, dndz=(z1, pz1),
                                      has_shear=False,
                                      ia_bias=(z1, a1))
    trc['i2'] = ccl.WeakLensingTracer(cosmo, dndz=(z2, pz2),
                                      has_shear=False,
                                      ia_bias=(z2, a2))
    trc['ct'] = ccl.CMBLensingTracer(cosmo, z_source=1100.)

    # Read benchmarks
    def read_bm(fname):
        th, xi = np.loadtxt(fname, unpack=True)
        return th, xi

    pre = dirdat + 'run_'
    post = nztyp + "_log_wt_"
    bms = {}
    theta, bms['dd_11'] = read_bm(pre + 'b1b1' + post + 'dd.txt')
    _, bms['dd_12'] = read_bm(pre + 'b1b2' + post + 'dd.txt')
    _, bms['dd_22'] = read_bm(pre + 'b2b2' + post + 'dd.txt')
    _, bms['dl_11'] = read_bm(pre + 'b1b1' + post + 'dl.txt')
    _, bms['dl_12'] = read_bm(pre + 'b1b2' + post + 'dl.txt')
    _, bms['dl_21'] = read_bm(pre + 'b2b1' + post + 'dl.txt')
    _, bms['dl_22'] = read_bm(pre + 'b2b2' + post + 'dl.txt')
    _, bms['di_11'] = read_bm(pre + 'b1b1' + post + 'di.txt')
    _, bms['di_12'] = read_bm(pre + 'b1b2' + post + 'di.txt')
    _, bms['di_21'] = read_bm(pre + 'b2b1' + post + 'di.txt')
    _, bms['di_22'] = read_bm(pre + 'b2b2' + post + 'di.txt')
    _, bms['ll_11_p'] = read_bm(pre + 'b1b1' + post + 'll_pp.txt')
    _, bms['ll_12_p'] = read_bm(pre + 'b1b2' + post + 'll_pp.txt')
    _, bms['ll_22_p'] = read_bm(pre + 'b2b2' + post + 'll_pp.txt')
    _, bms['ll_11_m'] = read_bm(pre + 'b1b1' + post + 'll_mm.txt')
    _, bms['ll_12_m'] = read_bm(pre + 'b1b2' + post + 'll_mm.txt')
    _, bms['ll_22_m'] = read_bm(pre + 'b2b2' + post + 'll_mm.txt')
    _, bms['li_11_p'] = read_bm(pre + 'b1b1' + post + 'li_pp.txt')
    _, bms['li_12_p'] = read_bm(pre + 'b1b2' + post + 'li_pp.txt')
    _, bms['li_22_p'] = read_bm(pre + 'b2b2' + post + 'li_pp.txt')
    _, bms['li_11_m'] = read_bm(pre + 'b1b1' + post + 'li_mm.txt')
    _, bms['li_12_m'] = read_bm(pre + 'b1b2' + post + 'li_mm.txt')
    _, bms['li_22_m'] = read_bm(pre + 'b2b2' + post + 'li_mm.txt')
    _, bms['ii_11_p'] = read_bm(pre + 'b1b1' + post + 'ii_pp.txt')
    _, bms['ii_12_p'] = read_bm(pre + 'b1b2' + post + 'ii_pp.txt')
    _, bms['ii_22_p'] = read_bm(pre + 'b2b2' + post + 'ii_pp.txt')
    _, bms['ii_11_m'] = read_bm(pre + 'b1b1' + post + 'ii_mm.txt')
    _, bms['ii_12_m'] = read_bm(pre + 'b1b2' + post + 'ii_mm.txt')
    _, bms['ii_22_m'] = read_bm(pre + 'b2b2' + post + 'ii_mm.txt')
    bms['theta'] = theta

    # Read error bars
    ers = {}
    d = np.loadtxt("benchmarks/data/sigma_clustering_Nbin5",
                   unpack=True)
    ers['dd_11'] = interp1d(d[0] / 60., d[1],
                            fill_value=d[1][0],
                            bounds_error=False)(theta)
    ers['dd_22'] = interp1d(d[0] / 60., d[2],
                            fill_value=d[2][0],
                            bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_ggl_Nbin5",
                   unpack=True)
    ers['dl_12'] = interp1d(d[0] / 60., d[1],
                            fill_value=d[1][0],
                            bounds_error=False)(theta)
    ers['dl_11'] = interp1d(d[0] / 60., d[2],
                            fill_value=d[2][0],
                            bounds_error=False)(theta)
    ers['dl_22'] = interp1d(d[0] / 60., d[3],
                            fill_value=d[3][0],
                            bounds_error=False)(theta)
    ers['dl_21'] = interp1d(d[0] / 60., d[4],
                            fill_value=d[4][0],
                            bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_xi+_Nbin5",
                   unpack=True)
    ers['ll_11_p'] = interp1d(d[0] / 60., d[1],
                              fill_value=d[1][0],
                              bounds_error=False)(theta)
    ers['ll_22_p'] = interp1d(d[0] / 60., d[2],
                              fill_value=d[2][0],
                              bounds_error=False)(theta)
    ers['ll_12_p'] = interp1d(d[0] / 60., d[3],
                              fill_value=d[3][0],
                              bounds_error=False)(theta)
    d = np.loadtxt("benchmarks/data/sigma_xi-_Nbin5",
                   unpack=True)
    ers['ll_11_m'] = interp1d(d[0] / 60., d[1],
                              fill_value=d[1][0],
                              bounds_error=False)(theta)
    ers['ll_22_m'] = interp1d(d[0] / 60., d[2],
                              fill_value=d[2][0],
                              bounds_error=False)(theta)
    ers['ll_12_m'] = interp1d(d[0] / 60., d[3],
                              fill_value=d[3][0],
                              bounds_error=False)(theta)
    print('setup time:', time.time() - t0)

    ccl.gsl_params.reload()
    return cosmo, trc, bms, ers, fl


# Commented out because we don't have this covariance
# ('g1', 'g2', 'dd_12', 'dd_12', 'gg', 1),
@pytest.mark.parametrize("t1,t2,bm,er,kind,pref",
                         [('g1', 'g1', 'dd_11', 'dd_11', 'NN', 1),
                          ('g2', 'g2', 'dd_22', 'dd_22', 'NN', 1),
                          ('g1', 'l1', 'dl_11', 'dl_11', 'NG', 1),
                          ('g1', 'l2', 'dl_12', 'dl_12', 'NG', 1),
                          ('g2', 'l1', 'dl_21', 'dl_21', 'NG', 1),
                          ('g2', 'l2', 'dl_22', 'dl_22', 'NG', 1),
                          ('g1', 'i1', 'di_11', 'dl_11', 'NG', 1),
                          ('g1', 'i2', 'di_12', 'dl_12', 'NG', 1),
                          ('g2', 'i1', 'di_21', 'dl_21', 'NG', 1),
                          ('g2', 'i2', 'di_22', 'dl_22', 'NG', 1),
                          ('l1', 'l1', 'll_11_p', 'll_11_p', 'GG+', 1),
                          ('l1', 'l2', 'll_12_p', 'll_12_p', 'GG+', 1),
                          ('l2', 'l2', 'll_22_p', 'll_22_p', 'GG+', 1),
                          ('l1', 'l1', 'll_11_m', 'll_11_m', 'GG-', 1),
                          ('l1', 'l2', 'll_12_m', 'll_12_m', 'GG-', 1),
                          ('l2', 'l2', 'll_22_m', 'll_22_m', 'GG-', 1),
                          ('i1', 'l1', 'li_11_p', 'll_11_p', 'GG+', 2),
                          ('i1', 'l2', 'li_12_p', 'll_11_p', 'GG+', 1),
                          ('i2', 'l2', 'li_22_p', 'll_22_p', 'GG+', 2),
                          ('i1', 'l1', 'li_11_m', 'll_11_m', 'GG-', 2),
                          ('i1', 'l2', 'li_12_m', 'll_11_m', 'GG-', 1),
                          ('i2', 'l2', 'li_22_m', 'll_22_m', 'GG-', 2),
                          ('i1', 'i1', 'ii_11_p', 'll_11_p', 'GG+', 1),
                          ('i1', 'i2', 'ii_12_p', 'll_12_p', 'GG+', 1),
                          ('i2', 'i2', 'ii_22_p', 'll_22_p', 'GG+', 1),
                          ('i1', 'i1', 'ii_11_m', 'll_11_m', 'GG-', 1),
                          ('i1', 'i2', 'ii_12_m', 'll_12_m', 'GG-', 1),
                          ('i2', 'i2', 'ii_22_m', 'll_22_m', 'GG-', 1)])
def test_xi(set_up, corr_method, t1, t2, bm, er, kind, pref):
    cosmo, trcs, bms, ers, fls = set_up
    method, errfac = corr_method

    global T0_CLS
    t0 = time.time()
    cl = ccl.angular_cl(cosmo, trcs[t1], trcs[t2], fls['ells'])
    T0_CLS += (time.time() - t0)

    ell = np.arange(fls['lmax'])
    cli = interp1d(fls['ells'], cl, kind='cubic')(ell)

    global T0
    t0 = time.time()
    xi = ccl.correlation(cosmo, ell=ell, C_ell=cli, theta=bms['theta'],
                         type=kind, method=method)
    T0 += (time.time() - t0)

    xi *= pref
    assert np.all(np.fabs(xi - bms[bm]) < ers[er] * errfac)



__all__ = ("angular_cl",)

import warnings

import numpy as np

from . import DEFAULT_POWER_SPECTRUM, CCLWarning, check, lib
from .pyutils import integ_types


def angular_cl(cosmo, tracer1, tracer2, ell, *,
               p_of_k_a=DEFAULT_POWER_SPECTRUM,
               l_limber=-1., limber_integration_method='qag_quad'):
    """Calculate the angular (cross-)power spectrum for a pair of tracers.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        tracer1 (:class:`~pyccl.tracers.Tracer`): a Tracer object,
            of any kind.
        tracer2 (:class:`~pyccl.tracers.Tracer`): a second Tracer object.
        ell (:obj:`float` or `array`): Angular multipole(s) at which to evaluate
            the angular power spectrum.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`, :obj:`str` or :obj:`None`): 3D Power
            spectrum to project. If a string, it must correspond to one of
            the non-linear power spectra stored in ``cosmo`` (e.g.
            ``'delta_matter:delta_matter'``).
        l_limber (:obj:`float`): Angular wavenumber beyond which Limber's
            approximation will be used.
        limber_integration_method (:obj:`str`): integration method to be used
            for the Limber integrals. Possibilities: ``'qag_quad'`` (GSL's
            `qag` method backed up by `quad` when it fails) and ``'spline'``
            (the integrand is splined and then integrated numerically).

    Returns:
        :obj:`float` or `array`: Angular (cross-)power spectrum values, \
            :math:`C_\\ell`, for the pair of tracers, as a function of \
            :math:`\\ell`.
    """ # noqa
    if cosmo['Omega_k'] != 0:
        warnings.warn(
            "CCL does not properly use the hyperspherical Bessel functions "
            "when computing angular power spectra in non-flat cosmologies!",
            category=CCLWarning)
    if limber_integration_method not in integ_types:
        raise ValueError(
            f"Unknown integration method {limber_integration_method}.")

    # we need the distances for the integrals
    cosmo.compute_distances()

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    psp = cosmo_in.parse_pk2d(p_of_k_a, is_linear=False)

    # Create tracer colections
    status = 0
    clt1, status = lib.cl_tracer_collection_t_new(status)
    clt2, status = lib.cl_tracer_collection_t_new(status)
    for t in tracer1._trc:
        status = lib.add_cl_tracer_to_collection(clt1, t, status)
    for t in tracer2._trc:
        status = lib.add_cl_tracer_to_collection(clt2, t, status)

    ell_use = np.atleast_1d(ell)

    # Check the values of ell are monotonically increasing
    if not (np.diff(ell_use) > 0).all():
        raise ValueError("ell values must be monotonically increasing")

    # Return Cl values, according to whether ell is an array or not
    cl, status = lib.angular_cl_vec(
        cosmo, clt1, clt2, psp, l_limber,
        ell_use, integ_types[limber_integration_method],
        ell_use.size, status)
    if np.ndim(ell) == 0:
        cl = cl[0]

    # Free up tracer collections
    lib.cl_tracer_collection_t_free(clt1)
    lib.cl_tracer_collection_t_free(clt2)

    check(status, cosmo=cosmo_in)
    return cl



import numpy as np
import pyccl as ccl

BEMBAR_TOLERANCE = 1e-3


def test_baccoemu_baryons():
    baryons = ccl.BaccoemuBaryons()
    cosmo = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    data = np.loadtxt("./benchmarks/data/baccoemu_baryons_fk.txt")

    k = data[:, 0] * cosmo['h']
    fk = data[:, 1]
    a = 1

    cclfk = baryons.boost_factor(cosmo, k, a)
    err = np.abs(fk / cclfk - 1)
    assert np.allclose(err, 0, atol=BEMBAR_TOLERANCE, rtol=0)


def test_baccoemu_baryons_A_s():
    baryons = ccl.BaccoemuBaryons()
    cosmo = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        A_s=2.2194e-9,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0)

    data = np.loadtxt("./benchmarks/data/baccoemu_baryons_fk.txt")

    k = data[:, 0] * cosmo['h']
    fk = data[:, 1]
    a = 1

    cclfk = baryons.boost_factor(cosmo, k, a)
    err = np.abs(fk / cclfk - 1)
    assert np.allclose(err, 0, atol=BEMBAR_TOLERANCE, rtol=0)



__all__ = ("FFTLogParams",)


class FFTLogParams:
    """Objects of this class store the FFTLog accuracy parameters.
    See documentation in :meth:`update_parameters` for a full
    description of all allowed parameters.
    """
    #: Anti-aliasing. Factor mulitplying the lower boundary.
    padding_lo_fftlog = 0.1
    #: Anti-aliasing. Factor mulitplying the upper boundary.
    padding_hi_fftlog = 10.

    #: Samples per decade for the Hankel transforms.
    n_per_decade = 100
    #: Extrapolation type (`linx_liny`, `linx_logy` etc.).
    extrapol = "linx_liny"

    #: Padding for intermediate transforms (lower bound).
    padding_lo_extra = 0.1
    #: Padding for intermediate transforms (upper bound).
    padding_hi_extra = 10.
    #: If True, high precision intermediate transforms.
    large_padding_2D = False

    #: Power law index used to prewhiten data before transform.
    plaw_fourier = -1.5
    #: Pre-whitening power law index for 2D and cumulative profiles.
    plaw_projected = -1.0

    @property
    def params(self):
        return ["padding_lo_fftlog", "padding_hi_fftlog", "n_per_decade",
                "extrapol", "padding_lo_extra", "padding_hi_extra",
                "large_padding_2D", "plaw_fourier", "plaw_projected"]

    def to_dict(self):
        """ Returns a dictionary containing this object's parameters.
        """
        return {param: getattr(self, param) for param in self.params}

    def __getitem__(self, name):
        return getattr(self, name)

    def __setattr__(self, name, value):
        raise AttributeError("FFTLogParams can only be updated via "
                             "`updated_parameters`.")

    def __repr__(self):
        return repr(self.to_dict())

    def __eq__(self, other):
        if self is other:
            True
        if type(self) is not type(other):
            return False
        return self.to_dict() == other.to_dict()

    def update_parameters(self, **kwargs):
        """Update the precision of FFTLog for the Hankel transforms.

        Arguments
        ---------
        padding_lo_fftlog: :obj:`float`
            Factor by which the minimum scale is multiplied to avoid
            aliasing. Default: 0.1.
        padding_hi_fftlog: :obj:`float`
            Factor by which the maximum scale is multiplied to avoid
            aliasing. Default: 10.
        n_per_decade : :obj:`float`
            Samples per decade for the Hankel transforms. Default: 100.
        extrapol : {'linx_liny', 'linx_logy'}
            Extrapolation type when FFTLog has narrower output support.
            Default ``'linx_liny'``.
        padding_lo_extra: :obj:`float`
            Additional minimum scale padding for double Hankel transforms,
            used when computing 2D projected and cumulative profiles. In
            these, the first transform goes from 3D real space to
            Fourier, and the second transform goes from Fourier to 2D
            real space.
            Default: 0.1.
        padding_hi_extra: :obj:`float`
            As ``padding_lo_extra`` for the maximum scale.
            Default: 10.
        large_padding_2D : :obj:`bool`
            Override ``padding_xx_extra`` in the intermediate transform,
            and use ``padding_xx_fftlog``. The default is False.
        plaw_fourier: :obj:`float`
            FFTLog pre-whitens its arguments (makes them flatter) to avoid
            aliasing. The ``plaw_fourier`` parameter describes the tilt of
            the profile, :math:`P(r) \\propto r^{\\mathrm{tilt}}`, for
            standard 3D transforms. Default: -1.5
        plaw_fourier_projected: :obj:`float`
            As ``plaw_fourier`` for 2D transforms (when computing 2D
            projected or cumulative profiles. Default: -1.0.
        """
        for name, value in kwargs.items():
            if name not in self.params:
                raise AttributeError(f"Parameter {name} does not exist.")
            object.__setattr__(self, name, value)



from ..halo_model_base import MassFunc, get_delta_c
from .angulo12 import *
from .bocquet16 import *
from .despali16 import *
from .jenkins01 import *
from .press74 import *
from .sheth99 import *
from .tinker08 import *
from .tinker10 import *
from .watson13 import *
from .bocquet20 import *



import numpy as np
import pytest
import pyccl as ccl
from pyccl import CCLWarning
from pyccl.pyutils import get_pk_spline_a, get_pk_spline_lk
from .test_cclobject import check_eq_repr_hash


def test_Pk2D_eq_repr_hash():
    # Test eq, repr, hash for Pk2D.
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function="bbks")
    cosmo.compute_linear_power()
    PK1 = cosmo.get_linear_power()
    PK2 = ccl.Pk2D.from_model(cosmo, "bbks")
    assert check_eq_repr_hash(PK1, PK2)
    assert check_eq_repr_hash(2*PK1, PK2, equal=False)

    # edge-case: same data
    empties = [ccl.Pk2D.__new__(ccl.Pk2D) for _ in range(2)]
    assert check_eq_repr_hash(*empties)
    assert check_eq_repr_hash(PK1, PK1)

    # edge-case: comparing different types
    assert check_eq_repr_hash(PK1, 1, equal=False)

    # edge-case: different extrapolation orders
    a_arr, lk_arr, pk_arr = PK1.get_spline_arrays()
    pk1 = ccl.Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                   is_logp=False, extrap_order_lok=0)
    pk2 = ccl.Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                   is_logp=False, extrap_order_lok=1)
    assert check_eq_repr_hash(pk1, pk2, equal=False)


def pk1d(k):
    return (k/0.1)**(-1)


def grw(a):
    return a


def pk2d(k, a):
    return pk1d(k)*grw(a)


def lpk2d(k, a):
    return np.log(pk2d(k, a))


def all_finite(vals):
    """
    Returns True if all elements are finite (i.e. not NaN or inf).
    """
    return np.all(np.isfinite(vals))


def test_pk2d_init():
    """
    Test initialization of Pk2D objects
    """

    # If no input
    with pytest.raises(ValueError):
        ccl.Pk2D()

    # Input function has incorrect signature
    with pytest.raises(TypeError):
        ccl.Pk2D.from_function(pk1d)

    # Input arrays have incorrect sizes
    lkarr = -4.+6*np.arange(100)/99.
    aarr = 0.05+0.95*np.arange(100)/99.
    pkarr = np.zeros([len(aarr), len(lkarr)])
    with pytest.raises(ValueError):
        ccl.Pk2D(a_arr=aarr, lk_arr=lkarr, pk_arr=pkarr[1:])

    # Scale factor is not monotonically increasing
    with pytest.raises(ValueError):
        ccl.Pk2D(a_arr=aarr[::-1], lk_arr=lkarr, pk_arr=pkarr)


def test_pk2d_smoke():
    """Make sure it works once."""
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=1e-10, n_s=0.96)
    lkarr = -4.+6*np.arange(100)/99.
    aarr = 0.05+0.95*np.arange(100)/99.
    pkarr = np.zeros([len(aarr), len(lkarr)])
    psp = ccl.Pk2D(a_arr=aarr, lk_arr=lkarr, pk_arr=pkarr)
    assert not np.isnan(psp(1E-2, 0.5, cosmo))


@pytest.mark.parametrize('model', ['bbks', 'eisenstein_hu',
                                   'eisenstein_hu_nowiggles'])
def test_pk2d_from_model(model):
    cosmo_fixed = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96)
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function=model)
    pk = ccl.Pk2D.from_model(cosmo_fixed, model=model)
    ks = np.geomspace(1E-3, 1E1, 128)
    for z in [0., 0.5, 2.]:
        a = 1./(1+z)
        pk1 = pk(ks, a, cosmo)
        pk2 = ccl.linear_matter_power(cosmo, ks, a)
        maxdiff = np.amax(np.fabs(pk1/pk2-1))
        assert maxdiff < 1E-10


@pytest.mark.parametrize('model', ['bbks', 'eisenstein_hu'])
def test_pk2d_from_model_fails(model):
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=1E-10, n_s=0.96,
        transfer_function='boltzmann_class')
    with pytest.raises(ccl.CCLError):
        ccl.Pk2D.from_model(cosmo, model=model)


def test_pk2d_from_model_raises():
    cosmo = ccl.CosmologyVanillaLCDM()
    with pytest.raises(ValueError):
        ccl.Pk2D.from_model(cosmo, model='bbkss')


def test_pk2d_function():
    """
    Test evaluation of Pk2D objects
    """

    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=1e-10, n_s=0.96)

    psp = ccl.Pk2D.from_function(pkfunc=lpk2d,
                                 spline_params=cosmo.cosmo.spline_params)
    psp2 = ccl.Pk2D.from_function(lpk2d)
    assert psp(1.0, 1.0) == psp2(1.0, 1.0)

    # Test at single point
    ktest = 1E-2
    atest = 0.5
    ptrue = pk2d(ktest, atest)
    phere = psp(ktest, atest, cosmo)
    assert np.allclose(phere, ptrue, atol=0, rtol=1e-7)
    dphere = psp(ktest, atest, cosmo, derivative=True)
    assert np.allclose(dphere, -1, atol=0, rtol=1e-6)

    ktest = 1
    atest = 0.5
    ptrue = pk2d(ktest, atest)
    phere = psp(ktest, atest, cosmo)
    assert np.allclose(phere, ptrue, atol=0, rtol=1e-7)
    dphere = psp(ktest, atest, cosmo, derivative=True)
    assert np.allclose(dphere, -1, atol=0, rtol=1e-6)

    # Test at array of points
    ktest = np.logspace(-3, 1, 10)
    ptrue = pk2d(ktest, atest)
    phere = psp(ktest, atest, cosmo)
    assert np.allclose(phere, ptrue, atol=0, rtol=1E-6)
    dphere = psp(ktest, atest, cosmo, derivative=True)
    assert np.allclose(dphere, -1, atol=0, rtol=1e-6)

    # Test input is not logarithmic
    psp = ccl.Pk2D.from_function(pkfunc=pk2d, is_logp=False,
                                 spline_params=cosmo.cosmo.spline_params)
    phere = psp(ktest, atest, cosmo)
    assert np.allclose(phere, ptrue, atol=0, rtol=1E-6)
    dphere = psp(ktest, atest, cosmo, derivative=True)
    assert np.allclose(dphere, -1, atol=0, rtol=1e-6)

    # Test input is arrays
    karr = np.logspace(-4, 2, 1000)
    aarr = np.linspace(0.01, 1., 100)
    parr = np.array([pk2d(karr, a) for a in aarr])
    psp = ccl.Pk2D(
        a_arr=aarr, lk_arr=np.log(karr), pk_arr=parr, is_logp=False)
    phere = psp(ktest, atest, cosmo)
    assert np.allclose(phere, ptrue, atol=0, rtol=1E-6)
    dphere = psp(ktest, atest, cosmo, derivative=True)
    assert np.allclose(dphere, -1.*np.ones_like(dphere), atol=0, rtol=1e-6)


def test_pk2d_from_function_spline_params():
    """Verify that passing spline_params as an argument works as expected."""
    k = np.logspace(-1, 0.5, 8)
    a = 0.8

    # Sampling from CCL's spline parameters.
    pk1 = ccl.Pk2D.from_function(pk2d, is_logp=False)
    a_arr, lk_arr, _ = pk1.get_spline_arrays()
    assert np.allclose(pk2d(k, a), pk1(k, a), atol=0, rtol=1e-10)
    assert np.array_equal(a_arr, get_pk_spline_a())
    assert np.array_equal(lk_arr, get_pk_spline_lk())

    # Sampling with custom spline parameters (C API).
    ccl.spline_params.N_K -= 10
    ccl.spline_params.A_SPLINE_NA_PK -= 10
    cosmo = ccl.CosmologyVanillaLCDM()  # contains a copy of the new params
    pk2 = ccl.Pk2D.from_function(pk2d, is_logp=False,
                                 spline_params=cosmo.cosmo.spline_params)
    a_arr, lk_arr, pk_arr = pk2.get_spline_arrays()
    assert np.allclose(pk2d(k, a), pk2(k, a), atol=0, rtol=1e-9)
    assert np.array_equal(a_arr, get_pk_spline_a())
    assert np.array_equal(lk_arr, get_pk_spline_lk())

    # Sampling with custom spline parameters (Python API)
    pk3 = ccl.Pk2D.from_function(pk2d, is_logp=False,
                                 spline_params=ccl.spline_params)
    a_arr2, lk_arr2, pk_arr2 = pk3.get_spline_arrays()
    assert np.array_equal(a_arr, a_arr2)
    assert np.array_equal(lk_arr, lk_arr2)
    assert np.array_equal(pk_arr, pk_arr2)

    ccl.spline_params.reload()


def test_pk2d_cells():
    """
    Test interplay between Pk2D and the Limber integrator
    """
    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=1e-10, n_s=0.96)
    z = np.linspace(0., 1., 200)
    n = np.exp(-((z-0.5)/0.1)**2)
    lens1 = ccl.WeakLensingTracer(cosmo, dndz=(z, n))
    ells = np.arange(2, 10)

    # Check that passing no power spectrum is fine
    cells = ccl.angular_cl(cosmo, lens1, lens1, ells)
    assert all_finite(cells)

    # Check that passing a bogus power spectrum fails as expected
    with pytest.raises(ValueError):
        ccl.angular_cl(cosmo, lens1, lens1, ells, p_of_k_a=1)

    # Check that passing a correct power spectrum runs as expected
    psp = ccl.Pk2D.from_function(lpk2d)
    cells = ccl.angular_cl(cosmo, lens1, lens1, ells, p_of_k_a=psp)
    assert all_finite(cells)

    ccl.gsl_params.reload()  # reset to the default parameters


def test_pk2d_parsing():
    a_arr = np.linspace(0.1, 1, 100)
    k_arr = np.geomspace(1E-4, 1E3, 1000)
    pk_arr = a_arr[:, None] * ((k_arr/0.01)/(1+(k_arr/0.01)**3))[None, :]

    psp = ccl.Pk2D(a_arr=a_arr, lk_arr=np.log(k_arr),
                   pk_arr=np.log(pk_arr))

    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    cosmo = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr,
                   'a:b': pk_arr})
    z = np.linspace(0., 1., 200)
    n = np.exp(-((z-0.5)/0.1)**2)
    lens1 = ccl.WeakLensingTracer(cosmo, dndz=(z, n))
    ells = np.linspace(2, 100, 10)

    cells1 = ccl.angular_cl(cosmo, lens1, lens1, ells,
                            p_of_k_a='delta_matter:delta_matter')
    cells2 = ccl.angular_cl(cosmo, lens1, lens1, ells,
                            p_of_k_a='a:b')
    cells3 = ccl.angular_cl(cosmo, lens1, lens1, ells,
                            p_of_k_a=psp)
    assert all_finite(cells1)
    assert all_finite(cells2)
    assert all_finite(cells3)
    assert np.all(np.fabs(cells1/cells2-1) < 1E-10)
    assert np.all(np.fabs(cells2/cells3-1) < 1E-10)

    # Wrong name
    with pytest.raises(KeyError):
        ccl.angular_cl(cosmo, lens1, lens1, ells,
                       p_of_k_a='a:c')

    # Wrong type
    with pytest.raises(ValueError):
        ccl.angular_cl(cosmo, lens1, lens1, ells,
                       p_of_k_a=3)

    ccl.gsl_params.reload()  # reset to the default parameters


def test_pk2d_get_spline_arrays():
    empty_pk2d = ccl.Pk2D.__new__(ccl.Pk2D)

    # Pk2D needs splines defined to get splines out
    with pytest.raises(ValueError):
        empty_pk2d.get_spline_arrays()


def test_pk2d_add():
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_a = np.outer(x, np.exp(log_y))
    zarr_b = np.outer(-1*x, 4*np.exp(log_y))

    empty_pk2d = ccl.Pk2D.__new__(ccl.Pk2D)
    pk2d_a = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=np.log(zarr_a),
                      is_logp=True)
    pk2d_b = ccl.Pk2D(a_arr=2*x, lk_arr=log_y, pk_arr=zarr_b,
                      is_logp=False)
    pk2d_b2 = ccl.Pk2D(a_arr=x, lk_arr=log_y+0.5, pk_arr=zarr_b,
                       is_logp=False)

    # This raises an error because the a ranges don't match
    with pytest.raises(ValueError):
        pk2d_a + pk2d_b
    # This raises an error because the k ranges don't match
    with pytest.raises(ValueError):
        pk2d_a + pk2d_b2
    # This raises an error because addition with an empty Pk2D should not work
    with pytest.raises(ValueError):
        pk2d_a + empty_pk2d

    pk2d_c = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=zarr_b,
                      is_logp=False)

    pk2d_d = pk2d_a + pk2d_c
    pk2d_d2 = pk2d_a + 1.0
    xarr_d, yarr_d, zarr_d = pk2d_d.get_spline_arrays()
    _, _, zarr_d2 = pk2d_d2.get_spline_arrays()

    assert np.allclose(x, xarr_d)
    assert np.allclose(log_y, yarr_d)
    assert np.allclose(zarr_a + zarr_b, zarr_d)
    assert np.allclose(zarr_a + 1.0, zarr_d2)

    pk2d_e = ccl.Pk2D(a_arr=x[1:-1], lk_arr=log_y[1:-1],
                      pk_arr=zarr_b[1:-1, 1:-1],
                      is_logp=False)

    # This raises a warning because the power spectra are not defined on the
    # same support
    with pytest.warns(CCLWarning):
        pk2d_f = pk2d_e + pk2d_a

    xarr_f, yarr_f, zarr_f = pk2d_f.get_spline_arrays()

    assert np.allclose((zarr_a + zarr_b)[1:-1, 1:-1], zarr_f)


def test_pk2d_mul_pow():
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_a = np.outer(x, np.exp(log_y))
    zarr_b = np.outer(-1*x, 4*np.exp(log_y))

    pk2d_a = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=np.log(zarr_a),
                      is_logp=True)
    pk2d_b = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=zarr_b,
                      is_logp=False)

    # This raises an error because multiplication is only defined for
    # float, int, and Pk2D
    with pytest.raises(TypeError):
        pk2d_a*np.array([0.1, 0.2])

    # This raises an error because exponention is only defined for
    # float and int
    with pytest.raises(TypeError):
        pk2d_a**pk2d_b

    # This raises a warning because the power spectrum is non-negative and the
    # power is non-integer
    with pytest.warns(CCLWarning):
        pk2d_b**0.5

    pk2d_g = pk2d_a * pk2d_b
    pk2d_h = 2*pk2d_a
    pk2d_i = pk2d_a**1.8

    _, _, zarr_g = pk2d_g.get_spline_arrays()
    _, _, zarr_h = pk2d_h.get_spline_arrays()
    _, _, zarr_i = pk2d_i.get_spline_arrays()

    assert np.allclose(zarr_a * zarr_b, zarr_g)
    assert np.allclose(2 * zarr_a, zarr_h)
    assert np.allclose(zarr_a**1.8, zarr_i)

    pk2d_j = (pk2d_a + 0.5*pk2d_i)**1.5
    _, _, zarr_j = pk2d_j.get_spline_arrays()
    assert np.allclose((zarr_a + 0.5*zarr_i)**1.5, zarr_j)


def test_pk2d_extrap_orders():
    # Check that setting extrap orders propagates down to the `psp`.
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_a = np.outer(x, np.exp(log_y))
    pk = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=np.log(zarr_a), is_logp=True)

    assert pk.extrap_order_hik == pk.psp.extrap_order_hik
    assert pk.extrap_order_lok == pk.psp.extrap_order_lok


def test_pk2d_eval_cosmo():
    # Check that `eval` can be called without `cosmo` and that an error
    # is raised when scale factor is out of interpolation range.
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function="bbks")
    cosmo.compute_linear_power()
    pk = cosmo.get_linear_power()
    assert pk(1., 1.) == pk(1., 1., cosmo)

    amin = pk.psp.amin
    pk(1., amin*0.99, cosmo)  # doesn't fail because cosmo is provided
    with pytest.raises(ValueError):
        pk(1., amin*0.99)


def test_pk2d_copy():
    # Check that copying works as intended (also check `bool`).
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_a = np.outer(x, np.exp(log_y))
    pk = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=np.log(zarr_a), is_logp=True)

    pkc = pk.copy()
    assert np.allclose(pk.get_spline_arrays()[-1],
                       pkc.get_spline_arrays()[-1],
                       rtol=1e-15)
    assert bool(pk) is bool(pkc) is True  # they both have `psp`

    pk = ccl.Pk2D.__new__(ccl.Pk2D)
    pkc = pk.copy()
    assert bool(pk) is bool(pkc) is False


def test_pk2d_operations():
    # Everything is based on the already tested `add`, `mul`, and `pow`,
    # so we don't need to test every accepted type separately.
    x = np.linspace(0.1, 1, 10)
    log_y = np.linspace(-3, 1, 20)
    zarr_a = np.outer(x, np.exp(log_y))
    pk0 = ccl.Pk2D(a_arr=x, lk_arr=log_y, pk_arr=np.log(zarr_a), is_logp=True)
    pk1, pk2 = pk0.copy(), pk0.copy()

    # sub, truediv
    assert np.allclose((pk1 - pk2).get_spline_arrays()[-1], 0, rtol=1e-15)
    assert np.allclose((pk1 / pk2).get_spline_arrays()[-1], 1, rtol=1e-15)

    # rsub, rtruediv
    assert np.allclose((1 - pk1).get_spline_arrays()[-1],
                       1 - pk1.get_spline_arrays()[-1])
    assert np.allclose((1 / pk1).get_spline_arrays()[-1],
                       1 / pk1.get_spline_arrays()[-1])

    # iadd, isub, imul, itruediv, ipow
    pk1 += pk1
    assert np.allclose((pk1 / pk2).get_spline_arrays()[-1], 2, rtol=1e-15)
    pk1 -= pk2
    assert np.allclose((pk1 / pk2).get_spline_arrays()[-1], 1, rtol=1e-15)
    pk1 *= pk1
    assert np.allclose(pk1.get_spline_arrays()[-1],
                       pk2.get_spline_arrays()[-1]**2,
                       rtol=1e-15)
    pk1 /= pk2
    assert np.allclose((pk1 / pk2).get_spline_arrays()[-1], 1, rtol=1e-15)
    pk1 **= 2
    assert np.allclose(pk1.get_spline_arrays()[-1],
                       pk2.get_spline_arrays()[-1]**2,
                       rtol=1e-15)


def test_pk2d_from_model_smoke():
    # Verify that both `from_model` methods are equivalent.
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function="bbks")
    pk1 = ccl.Pk2D.from_model(cosmo, "bbks")
    pk2 = cosmo.get_linear_power()
    assert np.all(pk1.get_spline_arrays()[-1] == pk2.get_spline_arrays()[-1])
